from __future__ import print_function
import nifty
import numpy



G = nifty.graph.UndirectedGraph
CG = G.EdgeContractionGraph

GMCO = G.MulticutObjective
CGMCO = CG.MulticutObjective



def testUndirectedGraph():

    g =  nifty.graph.UndirectedGraph(4)
    edges =  numpy.array([[0,1],[0,2],[0,3]],dtype='uint64')
    g.insertEdges(edges)

    edgeList = [e for e in g.edges()]
    assert edgeList == [0,1,2]

    nodeList = [e for e in g.nodes()]
    assert nodeList == [0,1,2,3]

    assert g.u(0) == 0
    assert g.v(0) == 1 
    assert g.u(1) == 0 
    assert g.v(1) == 2 
    assert g.u(2) == 0 
    assert g.v(2) == 3 


def testUndirectedGraphSerialization():

    gA =  nifty.graph.UndirectedGraph(4)
    edges =  numpy.array([[0,1],[0,2],[0,3]],dtype='uint64')
    gA.insertEdges(edges)

    serialization = gA.serialize()

    g = nifty.graph.UndirectedGraph()

    g.deserialize(serialization)

    edgeList = [e for e in g.edges()]
    assert edgeList == [0,1,2]

    nodeList = [e for e in g.nodes()]
    assert nodeList == [0,1,2,3]

    assert g.u(0) == 0
    assert g.v(0) == 1 
    assert g.u(1) == 0 
    assert g.v(1) == 2 
    assert g.u(2) == 0 
    assert g.v(2) == 3 








from __future__ import print_function

import os
import unittest
from shutil import rmtree

import numpy
import nifty
WITH_HDF5 = nifty.Configuration.WITH_HDF5

try:
    import h5py
    WITH_H5PY = True
except ImportError:
    WITH_H5PY= False


class TestHDF5(unittest.TestCase):
    tempFolder = './tmp_hdf5'

    def setUp(self):
        try:
            os.mkdir(self.tempFolder)
        except OSError:
            pass

    def tearDown(self):
        try:
            rmtree(self.tempFolder)
        except OSError:
            pass

    @unittest.skipUnless(WITH_HDF5 and WITH_H5PY,
                         "Need nifty-hdf5 and h5py")
    def test_hdf5_read_from_chunked(self):
        import nifty.hdf5 as nhdf5
        fpath = os.path.join(self.tempFolder, '_nifty_test_array__.h5')

        shape = (101, 102, 103)
        chunks = (10, 20, 30)
        data = numpy.ones(shape=shape, dtype='uint64')
        with h5py.File(fpath, 'a') as f:
            f.create_dataset("data", shape, dtype='uint64', data=data, chunks=chunks)

        hidT = nhdf5.openFile(fpath)
        array = nhdf5.Hdf5ArrayUInt64(hidT, "data")
        ashape = array.shape

        self.assertEqual(array.ndim, 3)
        self.assertEqual(len(ashape), 3)
        self.assertEqual(tuple(ashape), shape)

        self.assertTrue(array.isChunked)
        chunkShape = array.chunkShape
        self.assertEqual(len(chunkShape), 3)
        self.assertEqual(tuple(chunkShape), tuple(chunks))

        subarray = array[0:10, 0:10, 0:10]
        expected = data[0:10, 0:10, 0:10]
        self.assertEqual(subarray.shape, expected.shape)
        self.assertTrue(numpy.allclose(subarray, expected))

    @unittest.skipUnless(WITH_HDF5 and WITH_H5PY,
                         "Need nifty-hdf5 and h5py")
    def test_hdf5_read_from_non_chunked(self):
        import nifty.hdf5 as nhdf5
        fpath = os.path.join(self.tempFolder, '_nifty_test_array_.h5')

        shape = (101, 102, 103)
        data = numpy.ones(shape=shape, dtype='uint64')
        with h5py.File(fpath, 'a') as f:
            f.create_dataset("data", shape, dtype='uint64', data=data)

        hidT = nhdf5.openFile(fpath)
        array = nhdf5.Hdf5ArrayUInt64(hidT, "data")

        ashape = array.shape
        self.assertEqual(array.ndim, 3)
        self.assertEqual(len(ashape), 3)
        self.assertEqual(shape, tuple(ashape))

        self.assertFalse(array.isChunked)
        chunkShape = array.chunkShape
        self.assertEqual(len(chunkShape), 3)
        self.assertEqual(tuple(chunkShape), shape)

        subarray = array[0:10, 0:10, 0:10]
        expected = data[0:10, 0:10, 0:10]
        self.assertEqual(subarray.shape, expected.shape)
        self.assertTrue(numpy.allclose(subarray, expected))

    @unittest.skipUnless(WITH_HDF5, "Need nifty-hdf5")
    def test_create_chunked_array(self):
        import nifty.hdf5 as nhdf5
        fpath = os.path.join(self.tempFolder, '_nifty_test_array_.h5')

        hidT = nhdf5.createFile(fpath)
        shape = [101, 102, 103]
        chunks = [11, 12, 13]
        array = nhdf5.Hdf5ArrayUInt64(hidT, "data", shape, chunks)

        ashape = array.shape
        self.assertEqual(array.ndim, 3)
        self.assertEqual(ashape, shape)

        chunkShape = array.chunkShape
        self.assertEqual(chunkShape, chunks)

        ends = [10, 11, 12]

        toWrite = numpy.arange(ends[0]*ends[1]*ends[2]).reshape(ends)
        array[0:ends[0], 0:ends[1], 0:ends[2]] = toWrite
        subarray  = array[0:ends[0], 0:ends[1], 0:ends[2]]

        self.assertTrue(numpy.array_equal(toWrite, subarray))

    @unittest.skipUnless(WITH_HDF5, "Need nifty-hdf5")
    def test_create_zipped_array(self):
        import nifty.hdf5 as nhdf5
        fpath = os.path.join(self.tempFolder, '_nifty_test_array_.h5')

        shape = [101,102,103]
        chunks = [11,12,13]
        hidT = nhdf5.createFile(fpath)
        array = nhdf5.Hdf5ArrayUInt64(
            groupHandle=hidT,
            datasetName="data",
            shape=shape,
            chunkShape=chunks,
            compression=9
        )

        ashape = array.shape
        self.assertEqual(array.ndim, 3)
        self.assertEqual(shape, ashape)

        chunkShape = array.chunkShape
        self.assertEqual(chunkShape, chunks)

        ends = [10,11,12]

        toWrite = numpy.arange(ends[0]*ends[1]*ends[2]).reshape(ends)
        array[0:ends[0], 0:ends[1], 0:ends[2]] = toWrite
        subarray = array[0:ends[0], 0:ends[1], 0:ends[2]]

        self.assertTrue(numpy.array_equal(toWrite, subarray))


if __name__ == '__main__':
    unittest.main()



import unittest
import numpy
import os
from shutil import rmtree
import zipfile
from subprocess import call
import nifty.graph.rag as nrag
import nifty.hdf5 as nh5
import vigra
from functools import partial


class TestAccumulation(unittest.TestCase):

    @classmethod
    def setUpClass(self):

        urlData = 'https://www.dropbox.com/s/l1tgzlim8h1pb7w/test_data_anisotropic.zip?dl=0'
        call(['wget', '-Odata.zip', urlData])
        with zipfile.ZipFile('./data.zip') as f:
            f.extractall('.')
        os.remove('./data.zip')

        urlFeatures = 'https://www.dropbox.com/s/5nz0cjd9zdnfbmy/reference_features.zip?dl=0'
        call(['wget', '-Ofeatures.zip', urlFeatures])
        with zipfile.ZipFile('./features.zip') as f:
            f.extractall('./features')
        os.remove('./features.zip')

        self.nNodes = vigra.readHDF5('./data/seg.h5', 'data').max() + 1
        self.segFile = nh5.openFile('./data/seg.h5')
        self.segArray = nh5.hdf5Array('uint32', self.segFile, 'data')
        self.dataFile = nh5.openFile('./data/pmap.h5')
        self.dataArray = nh5.hdf5Array('float32', self.dataFile, 'data')

    @classmethod
    def tearDownClass(self):
        nh5.closeFile(self.segFile)
        nh5.closeFile(self.dataFile)
        rmtree('./data')
        rmtree('./features')

    def makeToyData(self):
        shape = (2, 10, 10)
        seg = numpy.zeros(shape, dtype='uint32')
        val = numpy.zeros(shape, dtype='float32')

        # the test segmentation
        seg[0, :5] = 0
        seg[0, 5:] = 1
        seg[1, :5] = 2
        seg[1, 5:] = 3

        # the test values
        val[0, :] = 0.
        val[1, :] = 1.
        return seg, val

    # check for the xy - edges is the same for all three accumulations
    def checkToyFeats(self, feats, zDir):
        self.assertEqual(feats[0, 0], 0.)
        self.assertEqual(feats[3, 0], 1.)
        if zDir == 0:
            self.assertEqual(feats[1, 0], .5)
            self.assertEqual(feats[2, 0], .5)
        elif zDir == 1:
            self.assertEqual(feats[1, 0], 0.)
            self.assertEqual(feats[2, 0], 0.)
        elif zDir == 2:
            self.assertEqual(feats[1, 0], 1.)
            self.assertEqual(feats[2, 0], 1.)

    def testFlatAccumulation(self):
        seg, val = self.makeToyData()
        rag = nrag.gridRag(seg, numberOfLabels=seg.max() + 1)

        # test the different z accumulations
        for zDir in (0, 1, 2):
            feats = nrag.accumulateEdgeFeaturesFlat(rag, val, val.min(), val.max(), zDir, 1)
            self.assertEqual(len(feats), rag.numberOfEdges)
            self.checkToyFeats(feats, zDir)

    def testReferenceFeatures(self):
        rag = nrag.gridRagStacked2DHdf5(self.segArray, self.nNodes)

        def singleFunctionTest(feature_function, name):

            def singleFeatureTest(fu, typ, zDir):
                xname = 'feats_%s_%s_%i_xy.h5' % (name, typ, zDir)
                zname = 'feats_%s_%s_%i_z.h5' % (name, typ, zDir)
                xy_file = nh5.createFile(xname)
                z_file = nh5.createFile(zname)
                xy_shape = [
                    rag.totalNumberOfInSliceEdges if typ in ('xy', 'both') else 1,
                    9 if name == 'standard' else 9 * 12
                ]
                xy_chunks = [min(2500, xy_shape[0]), xy_shape[1]]
                z_shape = [
                    rag.totalNumberOfInBetweenSliceEdges if typ in ('z', 'both') else 1,
                    9 if name == 'standard' else 9 * 12
                ]
                z_chunks = [min(2500, z_shape[0]), z_shape[1]]
                xy_array = nh5.hdf5Array('float32', xy_file, 'data', xy_shape, xy_chunks)
                z_array = nh5.hdf5Array('float32', z_file, 'data', z_shape, z_chunks)
                fu(rag, self.dataArray, xy_array, z_array, zDirection=zDir)
                xfeats = xy_array.readSubarray([0, 0], xy_shape)
                zfeats = z_array.readSubarray([0, 0], z_shape)
                nh5.closeFile(xy_file)
                nh5.closeFile(z_file)
                os.remove(xname)
                os.remove(zname)
                return xname, zname, xfeats, zfeats

            for typ in ('both', 'xy', 'z'):
                if typ == 'both':
                    new_fu = partial(feature_function, keepXYOnly=False, keepZOnly=False)
                elif typ == 'xy':
                    new_fu = partial(feature_function, keepXYOnly=True, keepZOnly=False)
                elif typ == 'z':
                    new_fu = partial(feature_function, keepXYOnly=False, keepZOnly=True)

                if typ == 'z':
                    for zDir in (0, 1, 2):
                        _, zname, _, zfeats = singleFeatureTest(new_fu, typ, zDir)
                        ref_feats = vigra.readHDF5(os.path.join('./features', zname), 'data')
                        self.assertTrue(numpy.allclose(zfeats, ref_feats))

                else:
                    zDir = 0
                    xname, zname, xfeats, zfeats = singleFeatureTest(new_fu, typ, zDir)
                    ref_feats_xy = vigra.readHDF5(os.path.join('./features', xname), 'data')
                    self.assertTrue(numpy.allclose(xfeats, ref_feats_xy))
                    if typ == 'both':
                        ref_feats_z = vigra.readHDF5(os.path.join('./features', zname), 'data')
                        self.assertTrue(numpy.allclose(zfeats, ref_feats_z))

        # standard feats
        singleFunctionTest(nrag.accumulateEdgeStandardFeatures, 'standard')
        # filter feats
        singleFunctionTest(nrag.accumulateEdgeFeaturesFromFilters, 'filter')

    def makeReferenceFeaturesStackedRag(self):
        rag = nrag.gridRagStacked2DHdf5(self.segArray, self.nNodes)

        def makeFeatures(feature_function, name):

            def singleFeature(fu, typ, zDir):
                xname = './feats_%s_%s_%i_xy.h5' % (name, typ, zDir)
                zname = './feats_%s_%s_%i_z.h5' % (name, typ, zDir)
                xy_file = nh5.createFile(xname)
                z_file = nh5.createFile(zname)
                xy_shape = [
                    rag.totalNumberOfInSliceEdges if typ in ('xy', 'both') else 1,
                    9 if name == 'standard' else 9 * 12
                ]
                xy_chunks = [min(2500, xy_shape[0]), xy_shape[1]]
                z_shape = [
                    rag.totalNumberOfInBetweenSliceEdges if typ in ('z', 'both') else 1,
                    9 if name == 'standard' else 9 * 12
                ]
                z_chunks = [min(2500, z_shape[0]), z_shape[1]]
                xy_array = nh5.hdf5Array('float32', xy_file, 'data', xy_shape, xy_chunks)
                z_array = nh5.hdf5Array('float32', z_file, 'data', z_shape, z_chunks)
                fu(rag, self.dataArray, xy_array, z_array, zDirection=zDir)
                nh5.closeFile(xy_file)
                nh5.closeFile(z_file)
                return xname, zname

            for typ in ('both', 'xy', 'z'):
                if typ == 'both':
                    new_fu = partial(feature_function, keepXYOnly=False, keepZOnly=False)
                elif typ == 'xy':
                    new_fu = partial(feature_function, keepXYOnly=True, keepZOnly=False)
                elif typ == 'z':
                    new_fu = partial(feature_function, keepXYOnly=False, keepZOnly=True)

                if typ == 'z':
                    for zDir in (0, 1, 2):
                        xname, _ = singleFeature(new_fu, typ, zDir)
                        os.remove(xname)

                else:
                    zDir = 0
                    _, zname = singleFeature(new_fu, typ, zDir)
                    if typ == 'xy':
                        os.remove(zname)

        # reference standard feats
        makeFeatures(nrag.accumulateEdgeStandardFeatures, 'standard')

        # reference filter feats
        makeFeatures(nrag.accumulateEdgeFeaturesFromFilters, 'filter')

    @unittest.expectedFailure
    def testStandardAccumulation(self):
        seg, val = self.makeToyData()
        rag = nrag.gridRagStacked2D(seg)

        # test the different z accumulations
        for zDir in (0, 1, 2):
            feats = nrag.accumulateEdgeStandardFeatures(rag,
                                                        val,
                                                        zDir)
            self.assertEqual(len(feats), rag.numberOfEdges)
            self.checkToyFeats(feats, zDir)


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import
from __future__ import print_function

from . import _tools as __tools
from ._tools import *

from concurrent.futures import ThreadPoolExecutor
from multiprocessing import cpu_count
import time
import sys
import threading
import random
import numpy


try:
    import progressbar as _progressbar
    _hasProgressbar = True
except ImportError:
    _hasProgressbar = False




__all__ = []

for key in __tools.__dict__.keys():
    __all__.append(key)
    try:
        __tools.__dict__[key].__module__='nifty.tools'
    except:
        pass





def rand_from_range_list(range_list, bad_slice_set):
    while(True):
        i = random.randint(0, len(range_list)-1)
        r = range_list[i]
        res =  random.randint(r[0],r[1])
        if res not in bad_slice_set:
            return res


# stupid impl....
# TODO make better
def rand_choice(probs, choose_from=None):
    probs = numpy.require(probs, dtype='float32')
    probs /= probs.sum()
    if choose_from is None:
       choose_from = range(len(probs))
    return numpy.random.choice(choose_from, 1, p=probs)[0]




def take(relabeling, toRelabel):
    shape = toRelabel.shape
    toRelabelFlat = toRelabel.ravel()
    return _tools._take(relabeling, toRelabelFlat).reshape(shape)


def mapFeaturesToLabelArray(label_array, feature_array,
                            ignore_label=-1,
                            fill_value=0.,
                            nb_threads=-1):
    ignore_label = -1 if ignore_label is None else ignore_label
    fill_value = 0. if fill_value is None else fill_value
    nb_threads = -1 if nb_threads is None else nb_threads
    shape = label_array.shape
    label_array_flat = label_array.ravel()
    return _tools._mapFeaturesToLabelArray(label_array_flat.astype(numpy.int64), feature_array.astype(numpy.float32), ignore_label, fill_value, nb_threads).reshape(shape + feature_array.shape[1:])


def takeDict(relabeling, toRelabel):
    shape = toRelabel.shape
    toRelabelFlat = toRelabel.ravel()
    return _tools._takeDict(relabeling, toRelabelFlat).reshape(shape)



def getSlicing(begin, end):
    return [slice(b,e) for b,e in zip(begin,end)]


def blocking(roiBegin, roiEnd, blockShape, blockShift=None):
    ndim = len(roiBegin)

    assert ndim == len(roiEnd)
    assert ndim == len(blockShape)
    if blockShift is not None:
        assert ndim == len(blockShift)
    else:
        blockShift = [0]*ndim


    if ndim == 1:
        blockingCls = Blocking1d
    elif ndim == 2:
        blockingCls = Blocking2d
    elif ndim == 3:
        blockingCls = Blocking3d
    elif ndim == 4:
        blockingCls = Blocking4d
    else:
        raise RuntimeError("only 1, 2, 3 and 4d blocking is currently implemented")

    return blockingCls(
        [int(v) for v in roiBegin],
        [int(v) for v in roiEnd],
        [int(v) for v in blockShape],
        [int(v) for v in blockShift]
    )



def parallelForEach(iterable, f, nWorkers=cpu_count() ,
                    showBar=False, size=None, name=None):
    if nWorkers == -1 or nWorkers is None:
        nWorkers = cpu_count()
    if not showBar:
        if nWorkers == 1 or nWorkers == 0:
            for i in iterable:
                f(i)
        else:
            with ThreadPoolExecutor(max_workers=nWorkers) as e:
                for i in iterable:
                    e.submit(f,i)

    else:
        if size is None:
            raise RuntimeError("if showBar==True, size must be specified")

        lock = threading.Lock()
        done = [0]

        with progressBar(size=size, name=name) as bar:
            def fTilde(val):

                f(val)

                with lock:
                    done[0] += 1
                    bar.update(done[0])
            parallelForEach(iterable=iterable,
                            f=fTilde,
                            nWorkers=nWorkers,
                            showBar=False,
                            size=None)





if not _hasProgressbar:

    class Progressbar:
        def __init__(self, maxValue, name=""):
            self.maxValue = maxValue

        def __enter__(self):
            #set things up
            return self

        def __exit__(self, type, value, traceback):

            infoStr = str(self.maxValue) + '/' + str(self.maxValue)
            sys.stdout.write('%s' % infoStr.ljust(20))
            pass
        def update(self, val):
            infoStr = str(val) + '/' + str(self.maxValue)
            sys.stdout.write('%s\r' % infoStr.ljust(20))
            sys.stdout.flush()
else:
    class Progressbar:
        def __init__(self, size, name=None):
            if name is None:
                name = ""
            widgets = [

                ' [',str(name), _progressbar.Timer(), ', ',_progressbar.Counter(),'/%s'%size,'] ',
                 _progressbar.Bar(),
                ' (', _progressbar.ETA(), ') ',
            ]

            self.bar = _progressbar.ProgressBar(
                maxval=size, widgets=widgets
            )

        def __enter__(self):
            self.bar.start()
            return self

        def __exit__(self, type, value, traceback):
            self.bar.finish()
            pass
        def update(self, val):
            self.bar.update(val)


def progressBar(size, name=None):
    return Progressbar(size,name)



import os
import unittest

import numpy
import nifty
import nifty.graph.rag as nrag

ON_WIN = os.name == 'nt'

try:
    import h5py
    WITH_H5PY = True
except ImportError:
    WITH_H5PY = False

try:
    from .test_rag import TestRagBase
except ImportError:
    from test_rag import TestRagBase


class TestRagStacked(TestRagBase):
    # shape of the small label array
    shape = [3, 2, 2]
    # small labels
    labels = numpy.array([[[0, 1],
                           [0, 1]],
                          [[2, 2],
                           [2, 3]],
                          [[4, 5],
                           [6, 6]]],
                         dtype='uint32')
    # graph edges for small labels
    shouldEdges = [(0, 1),
                   (0, 2),
                   (1, 2),
                   (1, 3),
                   (2, 3),
                   (2, 4),
                   (2, 5),
                   (2, 6),
                   (3, 6),
                   (4, 5),
                   (4, 6),
                   (5, 6)]
    # not graph edges for small labels
    shouldNotEdges = [(0, 3),
                      (0, 4),
                      (0, 5),
                      (0, 6),
                      (1, 4),
                      (1, 5),
                      (1, 6)]

    # shape of the big label array
    bigShape = [3, 4, 4]
    # big labels
    bigLabels = numpy.array([[[0, 0, 0, 0],
                              [1, 1, 1, 1],
                              [2, 2, 2, 2],
                              [2, 2, 2, 2]],
                             [[3, 3, 3, 3],
                              [3, 3, 3, 3],
                              [3, 3, 3, 3],
                              [3, 3, 3, 3]],
                             [[4, 4, 5, 5],
                              [4, 4, 5, 5],
                              [4, 4, 5, 5],
                              [4, 4, 5, 5]]],
                            dtype='uint32')
    # graph edges for big labels
    bigShouldEdges = [(0, 1),
                      (0, 3),
                      (1, 2),
                      (1, 3),
                      (2, 3),
                      (3, 4),
                      (3, 5),
                      (4, 5)]
    # not graph edges for big labels
    bigShouldNotEdges = [(0, 4),
                         (0, 5),
                         (1, 4),
                         (1, 5),
                         (2, 4),
                         (2, 5)]
    # graph edges for ignore graphs
    ignoreShouldEdges = [(1, 2),
                         (1, 3),
                         (2, 3),
                         (3, 4),
                         (3, 5),
                         (4, 5)]
    # not graph edges for ignore graphs
    ignoreShouldNotEdges = [(0, 1),
                            (0, 3),
                            (0, 4),
                            (0, 5),
                            (1, 4),
                            (1, 5),
                            (2, 4),
                            (2, 5)]

    def ignore_array_test(self, array, ragFunction):
        rag = ragFunction(array,
                          numberOfLabels=self.bigLabels.max() + 1,
                          ignoreLabel=0,
                          numberOfThreads=1)

        self.generic_rag_test(rag=rag,
                              numberOfNodes=self.bigLabels.max() + 1,
                              shouldEdges=self.ignoreShouldEdges,
                              shouldNotEdges=self.ignoreShouldNotEdges,
                              shape=self.bigShape)

    def test_grid_rag_stacked_ignore(self):
        array = numpy.zeros(self.bigShape, dtype='uint32')
        array[:] = self.bigLabels
        self.ignore_array_test(array, nrag.gridRagStacked2D)

    def small_array_test(self, array, ragFunction):
        rag = ragFunction(array,
                          numberOfLabels=self.labels.max() + 1,
                          numberOfThreads=-1)

        self.generic_rag_test(rag=rag,
                              numberOfNodes=self.labels.max() + 1,
                              shouldEdges=self.shouldEdges,
                              shouldNotEdges=self.shouldNotEdges,
                              shape=self.shape)

    def test_grid_rag_stacked2d(self):
        array = numpy.zeros(self.shape, dtype='uint32')
        array[:] = self.labels
        self.small_array_test(array, nrag.gridRagStacked2D)

    @unittest.skipUnless(nifty.Configuration.WITH_HDF5, "skipping hdf5 tests")
    def test_grid_rag_hdf5_stacked2d(self):
        import nifty.hdf5 as nhdf5
        hidT = nhdf5.createFile(self.path)
        chunkShape = [1, 2, 1]
        array = nhdf5.Hdf5ArrayUInt32(hidT, "data", self.shape, chunkShape)
        array[0:self.shape[0], 0:self.shape[1], 0:self.shape[2]] = self.labels
        self.small_array_test(array, nrag.gridRagStacked2DHdf5)
        nhdf5.closeFile(hidT)

    @unittest.skipIf(ON_WIN, "Fails on windows")
    @unittest.skipUnless(nifty.Configuration.WITH_Z5, "skipping z5 tests")
    def test_grid_rag_z5_stacked2d(self):
        import z5py
        import nifty.z5
        zfile = z5py.File(self.path, use_zarr_format=False)
        chunkShape = [1, 2, 1]
        array = zfile.create_dataset("data",
                                     dtype='uint32',
                                     shape=self.shape,
                                     chunks=chunkShape,
                                     compression='raw')
        array[:] = self.labels
        # we only pass the path and key to the dataset, because we
        # cannot properly link the python bindings for now
        # self.small_array_test(array, nrag.gridRagStacked2DZ5)
        self.small_array_test(nifty.z5.datasetWrapper('uint32', self.path, 'data'),
                              nrag.gridRagStacked2DZ5)

    def big_array_test(self, array, ragFunction):
        rag = ragFunction(array,
                          numberOfLabels=self.bigLabels.max() + 1,
                          numberOfThreads=-1)

        self.generic_rag_test(rag=rag,
                              numberOfNodes=self.bigLabels.max() + 1,
                              shouldEdges=self.bigShouldEdges,
                              shouldNotEdges=self.bigShouldNotEdges,
                              shape=self.bigShape)

    def test_grid_rag_stacked2d_large(self):
        array = numpy.zeros(self.bigShape, dtype='uint32')
        array[:] = self.bigLabels
        self.big_array_test(array, nrag.gridRagStacked2D)

    @unittest.skipUnless(nifty.Configuration.WITH_HDF5, "skipping hdf5 tests")
    def test_grid_rag_hdf5_stacked2d_large(self):
        import nifty.hdf5 as nhdf5
        hidT = nhdf5.createFile(self.path)
        chunkShape = [1, 2, 1]
        array = nhdf5.Hdf5ArrayUInt32(hidT, "data", self.bigShape, chunkShape)
        array[0:self.bigShape[0], 0:self.bigShape[1], 0:self.bigShape[2]] = self.bigLabels
        self.big_array_test(array, nrag.gridRagStacked2DHdf5)
        nhdf5.closeFile(hidT)

    @unittest.skipIf(ON_WIN, "Fails on windows")
    @unittest.skipUnless(nifty.Configuration.WITH_Z5, "skipping z5 tests")
    def test_grid_rag_z5_stacked2d_large(self):
        import z5py
        import nifty.z5
        zfile = z5py.File(self.path, use_zarr_format=False)
        chunkShape = [1, 2, 1]
        array = zfile.create_dataset("data",
                                     dtype='uint32',
                                     shape=self.bigShape,
                                     chunks=chunkShape,
                                     compression='raw')
        array[:] = self.bigLabels
        # we only pass the path and key to the dataset, because we
        # cannot properly link the python bindings for now
        # self.small_array_test(array, nrag.gridRagStacked2DZ5)
        self.big_array_test(nifty.z5.datasetWrapper('uint32', self.path, 'data'),
                            nrag.gridRagStacked2DZ5)

    def serialization_test(self, array, ragFunction):
        ragA = ragFunction(array,
                           numberOfLabels=self.bigLabels.max() + 1,
                           numberOfThreads=-1)

        self.generic_rag_test(rag=ragA,
                              numberOfNodes=self.bigLabels.max() + 1,
                              shouldEdges=self.bigShouldEdges,
                              shouldNotEdges=self.bigShouldNotEdges,
                              shape=self.bigShape)
        nrag.writeStackedRagToHdf5(ragA, self.path2)
        ragB = nrag.readStackedRagFromHdf5(labels=array,
                                           numberOfLabels=self.bigLabels.max() + 1,
                                           savePath=self.path2)
        self.generic_rag_test(rag=ragB,
                              numberOfNodes=self.bigLabels.max() + 1,
                              shouldEdges=self.bigShouldEdges,
                              shouldNotEdges=self.bigShouldNotEdges,
                              shape=self.bigShape)

        self.assertEqual(type(ragA), type(ragB))

        self.assertTrue((ragA.minMaxLabelPerSlice() == ragB.minMaxLabelPerSlice()).all())
        self.assertTrue((ragA.numberOfNodesPerSlice() == ragB.numberOfNodesPerSlice()).all())
        self.assertTrue((ragA.numberOfInSliceEdges() == ragB.numberOfInSliceEdges()).all())
        self.assertTrue((ragA.numberOfInBetweenSliceEdges() == ragB.numberOfInBetweenSliceEdges()).all())
        self.assertTrue((ragA.inSliceEdgeOffset() == ragB.inSliceEdgeOffset()).all())
        self.assertTrue((ragA.betweenSliceEdgeOffset() == ragB.betweenSliceEdgeOffset()).all())

    @unittest.skipUnless(WITH_H5PY, "skipping explicit serialization tests")
    def test_stacked_rag_serialize_deserialize(self):
        array = numpy.zeros(self.bigShape, dtype='uint32')
        array[:] = self.bigLabels
        self.serialization_test(array, nrag.gridRagStacked2D)

    @unittest.skipUnless(nifty.Configuration.WITH_HDF5 and WITH_H5PY,
                         "skipping hdf5 serialization tests")
    def test_stacked_rag_hdf5_serialize_deserialize(self):
        import nifty.hdf5 as nhdf5
        hidT = nhdf5.createFile(self.path)
        chunkShape = [1, 2, 1]
        array = nhdf5.Hdf5ArrayUInt32(hidT, "data", self.bigShape, chunkShape)
        array[0:self.bigShape[0], 0:self.bigShape[1], 0:self.bigShape[2]] = self.bigLabels
        self.serialization_test(array, nrag.gridRagStacked2DHdf5)
        nhdf5.closeFile(hidT)

    @unittest.skipIf(ON_WIN, "Fails on windows")
    @unittest.skipUnless(nifty.Configuration.WITH_Z5 and WITH_H5PY,
                         "skipping z5 serialization tests")
    def test_grid_rag_z5_serialize_deserialize(self):
        import z5py
        import nifty.z5
        zfile = z5py.File(self.path, use_zarr_format=False)
        chunkShape = [1, 2, 1]
        array = zfile.create_dataset("data",
                                     dtype='uint32',
                                     shape=self.bigShape,
                                     chunks=chunkShape,
                                     compression='raw')
        array[:] = self.bigLabels
        # we only pass the path and key to the dataset, because we
        # cannot properly link the python bindings for now
        self.serialization_test(nifty.z5.datasetWrapper('uint32', self.path, "data"),
                                nrag.gridRagStacked2DZ5)


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np
import nifty.tools as nt


class TestMergeHelper(unittest.TestCase):
    uv_ids = np.array([[0, 1],
                       [0, 2],
                       [0, 3],
                       [0, 1],
                       [0, 2],
                       [0, 3],
                       [1, 2]], dtype='uint64')
    indicator = np.array([1,
                          0,
                          1,
                          1,
                          1,
                          0,
                          1], dtype='uint64')

    def test_compute_votes(self):
        sizes = np.zeros_like(self.indicator, dtype='uint64')
        uv_ids_merged, indicators_merged = nt.computeMergeVotes(self.uv_ids, self.indicator, sizes, False)
        unique_ids = np.unique(self.uv_ids, axis=0)
        self.assertEqual(len(uv_ids_merged), len(unique_ids))
        self.assertTrue((uv_ids_merged == unique_ids).all())
        self.assertEqual(len(indicators_merged), len(unique_ids))

        expected_indicator_nom = np.array([2, 1, 1, 1])
        expected_indicator_den = np.array([2, 2, 2, 1])

        self.assertTrue((indicators_merged[:, 0] == expected_indicator_nom).all())
        self.assertTrue((indicators_merged[:, 1] == expected_indicator_den).all())

    def test_merge_votes(self):
        sizes = np.zeros_like(self.indicator, dtype='uint64')
        uv_ids_merged_a, indicators_merged_a = nt.computeMergeVotes(self.uv_ids, self.indicator, sizes, False)
        uv_ids_merged_b, indicators_merged_b = nt.computeMergeVotes(self.uv_ids, self.indicator, sizes, False)

        uv_ids_merged, indicators_merged = nt.mergeMergeVotes(np.concatenate([uv_ids_merged_a,
                                                                              uv_ids_merged_b], axis=0),
                                                             np.concatenate([indicators_merged_a,
                                                                             indicators_merged_b], axis=0))

        unique_ids = np.unique(self.uv_ids, axis=0)
        self.assertEqual(len(uv_ids_merged), len(unique_ids))
        self.assertTrue((uv_ids_merged == unique_ids).all())
        self.assertEqual(len(indicators_merged), len(unique_ids))

        expected_indicator_nom = 2 * np.array([2, 1, 1, 1])
        expected_indicator_den = 2 * np.array([2, 2, 2, 1])

        self.assertTrue((indicators_merged[:, 0] == expected_indicator_nom).all())
        self.assertTrue((indicators_merged[:, 1] == expected_indicator_den).all())


if __name__ == '__main__':
    unittest.main()



import unittest
import random
import numpy

import nifty
import nifty.graph
import nifty.graph.opt.mincut


class TestMincutSolver(unittest.TestCase):
    def setUp(self):
        numpy.random.seed(7)

    def generateGrid(self, gridSize):
        def nid(x, y):
            return x*gridSize[1] + y
        G = nifty.graph.UndirectedGraph
        g =  G(gridSize[0] * gridSize[1])
        for x in range(gridSize[0]):
            for y in range(gridSize[1]):

                u = nid(x,y)

                if x + 1 < gridSize[0]:

                    v = nid(x+1, y)
                    g.insertEdge(u, v)

                if y + 1 < gridSize[1]:

                    v = nid(x, y+1)
                    g.insertEdge(u, v)

        return g, nid

    def gridModel(self, gridSize = [5,5],  weightRange = [-2,1]):
        g,nid = self.generateGrid(gridSize)
        d = weightRange[1] - weightRange[0]
        w = numpy.random.rand(g.numberOfEdges)*d + float(weightRange[0])
        print(w.min(),w.max())
        obj = nifty.graph.opt.mincut.mincutObjective(g,w)
        return obj

    @unittest.skipUnless(nifty.Configuration.WITH_QPBO, "need qpbo")
    def testMincutQpbo(self):
        objective = self.gridModel(gridSize=[4,4])
        solver = objective.mincutQpboFactory(improve=False).create(objective)
        visitor = objective.verboseVisitor(1)
        arg = solver.optimize(visitor)

    @unittest.skipUnless(nifty.Configuration.WITH_QPBO, "need qpbo")
    def testMincutQpboImprove(self):
        objective = self.gridModel(gridSize=[8,8])
        solver = objective.mincutQpboFactory(improve=True).create(objective)
        visitor = objective.verboseVisitor(1)
        arg = numpy.zeros(4,dtype='uint8')
        arg = solver.optimize(visitor)


if __name__ == '__main__':
    unittest.main()



from __future__ import print_function
import unittest
import nifty
import numpy


class TestLiftedMulticutObjective(unittest.TestCase):
    def generateGrid(self, gridSize):
        def nid(x, y):
            return x*gridSize[1] + y
        G = nifty.graph.UndirectedGraph
        g =  G(gridSize[0] * gridSize[1])
        for x in range(gridSize[0]):
            for y in range(gridSize[1]):

                u = nid(x,y)

                if x + 1 < gridSize[0]:

                    v = nid(x+1, y)
                    g.insertEdge(u, v)

                if y + 1 < gridSize[1]:

                    v = nid(x, y+1)
                    g.insertEdge(u, v)

        return g, nid

    def testInsertLiftedEdgesBfsSmall(self):
        g,nid = self.generateGrid([2,2])


        obj = nifty.graph.lifted_multicut.liftedMulticutObjective(g)
        graph = obj.graph
        liftedGraph = obj.liftedGraph


        self.assertEqual(graph.numberOfEdges,g.numberOfEdges)
        self.assertEqual(liftedGraph.numberOfNodes, graph.numberOfNodes)
        self.assertEqual(liftedGraph.numberOfEdges, graph.numberOfEdges)

        # this should not add any edge at all
        distance = obj.insertLiftedEdgesBfs(1, returnDistance=True)
        postEdges = liftedGraph.numberOfEdges
        self.assertEqual(liftedGraph.numberOfEdges, graph.numberOfEdges)
        self.assertEqual(len(distance),0)

        # this should  add edges
        distance = obj.insertLiftedEdgesBfs(2, returnDistance=True)
        postEdges = liftedGraph.numberOfEdges
        self.assertEqual(liftedGraph.numberOfEdges, 6)
        self.assertEqual(len(distance),2)
        self.assertEqual(distance[0],2)
        self.assertEqual(distance[1],2)

    def testInsertLiftedEdgesBfsBig(self):

        g,nid = self.generateGrid([10, 10])
        obj = nifty.graph.lifted_multicut.liftedMulticutObjective(g)
        graph = obj.graph
        liftedGraph = obj.liftedGraph

        self.assertEqual(graph.numberOfEdges,g.numberOfEdges)
        self.assertEqual(liftedGraph.numberOfNodes, graph.numberOfNodes)
        self.assertEqual(liftedGraph.numberOfEdges, graph.numberOfEdges)

        # this should not add any edge at all
        obj.insertLiftedEdgesBfs(1)
        postEdges = liftedGraph.numberOfEdges
        self.assertEqual(liftedGraph.numberOfEdges, graph.numberOfEdges)

        # this should  add edges
        obj.insertLiftedEdgesBfs(2)
        postEdges = liftedGraph.numberOfEdges

        # check a few lifted edges
        node = nid(0,0)
        self.assertNotEqual( liftedGraph.findEdge(node, nid(0,1)) , -1 )
        self.assertNotEqual( liftedGraph.findEdge(node, nid(0,2)) , -1 )
        self.assertEqual( liftedGraph.findEdge(node, nid(0,3))    , -1 )


class TestLiftedMulticutGridGraphObjective(unittest.TestCase):
    def generateGrid(self, gridSize):
        def nid(x, y):
            return x*gridSize[1] + y
        g = nifty.graph.undirectedGridGraph(gridSize)

        return g, nid

    def testInsertLiftedEdgesBfsSmall(self):
        g,nid = self.generateGrid([2,2])

        obj = nifty.graph.lifted_multicut.liftedMulticutObjective(g)
        graph = obj.graph
        liftedGraph = obj.liftedGraph

        self.assertEqual(graph.numberOfEdges,g.numberOfEdges)
        self.assertEqual(liftedGraph.numberOfNodes, graph.numberOfNodes)
        self.assertEqual(liftedGraph.numberOfEdges, graph.numberOfEdges)

        # this should not add any edge at all
        distance = obj.insertLiftedEdgesBfs(1, returnDistance=True)
        postEdges = liftedGraph.numberOfEdges
        self.assertEqual(liftedGraph.numberOfEdges, graph.numberOfEdges)
        self.assertEqual(len(distance),0)

        # this should  add edges
        distance = obj.insertLiftedEdgesBfs(2, returnDistance=True)
        postEdges = liftedGraph.numberOfEdges
        self.assertEqual(liftedGraph.numberOfEdges, 6)
        self.assertEqual(len(distance),2)
        self.assertEqual(distance[0],2)
        self.assertEqual(distance[1],2)

    def testInsertLiftedEdgesBfsBig(self):

        g,nid = self.generateGrid([10, 10])
        obj = nifty.graph.lifted_multicut.liftedMulticutObjective(g)
        graph = obj.graph
        liftedGraph = obj.liftedGraph

        self.assertEqual(graph.numberOfEdges,g.numberOfEdges)
        self.assertEqual(liftedGraph.numberOfNodes, graph.numberOfNodes)
        self.assertEqual(liftedGraph.numberOfEdges, graph.numberOfEdges)

        # this should not add any edge at all
        obj.insertLiftedEdgesBfs(1)
        postEdges = liftedGraph.numberOfEdges
        self.assertEqual(liftedGraph.numberOfEdges, graph.numberOfEdges)

        # this should  add edges
        obj.insertLiftedEdgesBfs(2)
        postEdges = liftedGraph.numberOfEdges

        # check a few lifted edges
        node = nid(0,0)
        self.assertNotEqual( liftedGraph.findEdge(node, nid(0,1)) , -1 )
        self.assertNotEqual( liftedGraph.findEdge(node, nid(0,2)) , -1 )
        self.assertEqual( liftedGraph.findEdge(node, nid(0,3))    , -1 )


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import
from . import _long_range_adjacency as __long_range_adjacency
from ._long_range_adjacency import *
from .. import Configuration
from ..rag import gridRag

__all__ = []
for key in __long_range_adjacency.__dict__.keys():
    try:
        __long_range_adjacency.__dict__[key].__module__ = 'nifty.graph.long_range_adjacency'
    except:
        pass

    __all__.append(key)


def longRangeAdjacency(labels, numberOfLabels=None, longRange=None,
                       ignoreLabel=False, numberOfThreads=-1, serialization=None):
    assert labels.ndim == 3
    if serialization is None:
        assert numberOfLabels is not None and longRange is not None
        return explicitLabelsLongRangeAdjacency(labels, longRange, numberOfLabels, ignoreLabel, numberOfThreads)
    else:
        return explicitLabelsLongRangeAdjacency(labels, serialization)


if Configuration.WITH_HDF5:

    def longRangeAdjacencyHDF5(labels, numberOfLabels=None, longRange=None,
                               ignoreLabel=False, numberOfThreads=-1, serialization=None):
        if serialization is None:
            return hdf5LabelsLongRangeAdjacency(labels,  longRange, numberOfLabels, ignoreLabel, numberOfThreads)
        else:
            assert numberOfLabels is not None and longRange is not None
            return hdf5LabelsLongRangeAdjacency(labels, serialization)





from __future__ import absolute_import
from .. import Configuration

__all__ = []

if Configuration.WITH_HDF5:
    from ._hdf5 import *
    from . import _hdf5 as __hdf5
    for key in __hdf5.__dict__.keys():
        __all__.append(key)
        try:
            __hdf5.__dict__[key].__module__='nifty.hdf5'
        except:
            pass
else:
    pass

import math
import numpy
import numbers

if Configuration.WITH_HDF5:


    # convenience wrapper around hdf5 array
    def hdf5Array(dtype, *args, **kwargs):

        if numpy.dtype(dtype) == numpy.dtype("uint8"):
            return Hdf5ArrayUInt8(*args, **kwargs)
        elif numpy.dtype(dtype) == numpy.dtype("uint16"):
            return Hdf5ArrayUInt16(*args, **kwargs)
        elif numpy.dtype(dtype) == numpy.dtype("uint32"):
            return Hdf5ArrayUInt32(*args, **kwargs)
        elif numpy.dtype(dtype) == numpy.dtype("uint64"):
            return Hdf5ArrayUInt64(*args, **kwargs)

        elif numpy.dtype(dtype) == numpy.dtype("int8"):
            return Hdf5ArrayInt8(*args, **kwargs)
        elif numpy.dtype(dtype) == numpy.dtype("int16"):
            return Hdf5ArrayInt16(*args, **kwargs)
        elif numpy.dtype(dtype) == numpy.dtype("int32"):
            return Hdf5ArrayInt32(*args, **kwargs)
        elif numpy.dtype(dtype) == numpy.dtype("int64"):
            return Hdf5ArrayInt64(*args, **kwargs)

        elif numpy.dtype(dtype) == numpy.dtype("float32"):
            return Hdf5ArrayFloat32(*args, **kwargs)
        elif numpy.dtype(dtype) == numpy.dtype("float64"):
            return Hdf5ArrayFloat64(*args, **kwargs)

        else:
            raise RuntimeError("Datatype %s not supported!" % (str(dtype),))


    def unionFindWatershed(heightMapFilename,
                           heightMapDataset,
                           labelsFilename,
                           labelsDataset,
                           blockShape,
                           numberOfThreads=-1):

        nDim  =  len(blockShape)
        # pow2 blockshape
        #
        #pow2Blockshape = [2**(math.ceil(math.log(v)/math.log(2))) for v in blockShape]
        blockShapeArray = [128] * nDim

        fname = "blockwiseWatershed_float32_uint32_%dd"%nDim

        f = _hdf5.__dict__[fname]

        f(heightMapFilename, heightMapDataset,
          labelsFilename, labelsDataset,
          list(blockShape), list(blockShapeArray),int(numberOfThreads))






    def __extendHdf5Array():
        hdf5Arrays = [
            Hdf5ArrayUInt8,
            Hdf5ArrayUInt16,
            Hdf5ArrayUInt32,
            Hdf5ArrayUInt64,
            Hdf5ArrayInt8,
            Hdf5ArrayInt16,
            Hdf5ArrayInt32,
            Hdf5ArrayInt64,
            Hdf5ArrayFloat32,
            Hdf5ArrayFloat64
        ]

        def getItem(self, indexing):
            dim = self.ndim
            roiBegin = [None]*dim
            roiEnd = [None]*dim
            for d in range(dim):
                indexingObj = indexing[d]
                if isinstance(indexingObj, numbers.Integral):
                    roiBegin[d] = int(indexingObj)
                    roiEnd[d] = int(indexingObj+1)
                else:
                    roiBegin[d] = int(indexingObj.start)
                    roiEnd[d] = int(indexingObj.stop)
                    step = indexingObj.step
                    if step is not None and  step != 1:
                        raise RuntimeError("currently step must be 1 in slicing but step is %d"%sliceObj.step)


            return self.readSubarray(roiBegin, roiEnd)


        def setItem(self, slicing, value):
            asArray = numpy.require(value)

            dim = self.ndim
            roiBegin = [None]*dim
            roiEnd = [None]*dim
            shape = [None]*dim
            for d in range(dim):
                sliceObj = slicing[d]
                roiBegin[d] = int(sliceObj.start)
                roiEnd[d] = int(sliceObj.stop)
                if roiEnd[d] - roiBegin[d] != asArray.shape[d]:
                    raise RuntimeError("array to write does not match slicing shape")
                step = sliceObj.step
                if step is not None and  step != 1:
                    raise RuntimeError("currently step must be 1 in slicing but step is %d"%sliceObj.step)

            return self.writeSubarray(roiBegin, asArray)


        for array in hdf5Arrays:
            array.__getitem__ = getItem
            array.__setitem__ = setItem


    __extendHdf5Array()
    del __extendHdf5Array



import unittest
import numpy as np
import nifty.graph.rag as nrag



class TestCarving(unittest.TestCase):

    def make_labels(self, ndim):
        shape = ndim * (100,)
        x = np.random.randint(0, 100, size=shape,
                              dtype='uint32')
        return x

    def _test_carving(self, ndim):
        from nifty.carving import carvingSegmenter
        x = self.make_labels(ndim)
        rag = nrag.gridRag(x, numberOfLabels=int(x.max()) + 1)
        edgeWeights = 128 * np.random.rand(rag.numberOfEdges).astype('float32')
        segmenter = carvingSegmenter(rag, edgeWeights)

        noBoundaryBelow = 64.
        for bias in (.9, .95, 1.):
            seeds = np.random.randint(0, 3, size=rag.numberOfNodes)
            out = segmenter(seeds, bias, noBoundaryBelow)
            self.assertEqual(len(out), rag.numberOfNodes)
            self.assertTrue(np.allclose(np.unique(out), [1, 2]))

    def test_carving_2d(self):
        self._test_carving(2)

    def test_carving_3d(self):
        self._test_carving(3)


if __name__ == '__main__':
    unittest.main()



import unittest

from nifty.graph import accumulate_affinities_mean_and_length
import numpy as np


class TestAffinityAccumulation(unittest.TestCase):
    IMAGE_SHAPE = (10, 10, 10)

    def test_accumulate_affinity_mean_and_length(self):
        offsets = [
            # Direct 3D neighborhood:
            [-1, 0, 0], [0, -1, 0], [0, 0, -1],
            # Long-range connections:
            [-1, -1, -1]]

        # Generate some random affinities:
        random_affinities = np.random.uniform(size=(len(offsets),) + self.IMAGE_SHAPE).astype('float32')
        random_labels = np.random.randint(25, size=self.IMAGE_SHAPE)

        mean, count = accumulate_affinities_mean_and_length(random_affinities, offsets, random_labels,
                                                                 offset_weights=[2, 4, 1, 5])

    def test_accumulate_affinity_mean_and_length_inside_cluster(self):
        offsets = [
            # Direct 3D neighborhood:
            [-1, 0, 0], [0, -1, 0], [0, 0, -1],
            # Long-range connections:
            [-1, -1, -1]]

        # Generate some random affinities:
        random_affinities = np.random.uniform(size=(len(offsets),) + self.IMAGE_SHAPE).astype('float32')
        random_labels = np.random.randint(25, size=self.IMAGE_SHAPE)

        mean, count = accumulate_affinities_mean_and_length(
            random_affinities, offsets, random_labels,
            offset_weights=[2, 4, 1, 5])


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import,print_function
from .. import graph
from .. import filters

from skimage.feature import peak_local_max as __peak_local_max
import skimage.segmentation
from skimage.transform import resize as __imresize
from scipy.ndimage import zoom as __zoom
import scipy.ndimage
from matplotlib.colors import ListedColormap as __ListedColormap
import numpy



def slic(image, nSegments, components):
    """ same as skimage.segmentation.slic """
    return skimage.segmentation.slic(image, n_segments=nSegments,
        compactness=compactness)


def seededWatersheds(heightMap, seeds=None, method="node_weighted", acc="max"):
    """Seeded watersheds segmentation

    Get a segmentation via seeded watersheds.
    This is a high level wrapper around
    :func:`nifty.graph.nodeWeightedWatershedsSegmentation`
    and :func:`nifty.graph.nodeWeightedWatershedsSegmentation`.


    Args:
        heightMap (numpy.ndarray) : height / evaluation map
        seeds (numpy.ndarray) : Seeds as non zero elements in the array.
            (default: {nifty.segmentation.localMinimaSeeds(heightMap)})
        method (str): Algorithm type can be:

            *   "node_weighted": ordinary node weighted watershed
            *   "edge_weighted": edge weighted watershed (minimum spanning tree)

            (default: {"max"})

        acc (str): If method is "edge_weighted", one needs to specify how
            to convert the heightMap into an edgeMap.
            This parameter specificities this method.
            Allow values are:

            *   'min' : Take the minimum value of the endpoints of the edge
            *   'max' : Take the minimum value of the endpoints of the edge
            *   'sum' : Take the sum  of the values of the endpoints of the edge
            *   'prod' : Take the product of the values of the endpoints of the edge
            *   'interpixel' : Take the value of the image at the interpixel
                coordinate in between the two endpoints.
                To do this the image is resampled to have shape :math: `2 \cdot shape -1 `

            (default: {"max"})

    Returns:
        numpy.ndarray : the segmentation

    Raises:
        RuntimeError: [description]
    """
    if seeds is None:
        seeds = localMinimaSeeds(heightMap)

    hshape = heightMap.shape
    sshape = seeds.shape
    shape = sshape


    ishape = [2*s -1 for s in shape]
    gridGraph = graph.gridGraph(shape)

    # node watershed
    if method == "node_weighted":
        assert hshape == sshape
        seg = graph.nodeWeightedWatershedsSegmentation(graph=gridGraph, seeds=seeds.ravel(),
                nodeWeights=heightMap.ravel())
        seg = seg.reshape(shape)

    elif method == "edge_weighted":
        if acc != 'interpixel':
            assert hshape == sshape
            gridGraphEdgeStrength = gridGraph.imageToEdgeMap(heightMap, mode=acc)

        else:
            if(hshape == shape):
                iHeightMap = __imresize(heightMap, ishape, interp='bicubic')
                gridGraphEdgeStrength = gridGraph.imageToEdgeMap(iHeightMap, mode=acc)
            elif(hshape == ishape):
                gridGraphEdgeStrength = gridGraph.imageToEdgeMap(heightMap, mode=acc)
            else:
                raise RuntimeError("height map has wrong shape")

        seg = graph.edgeWeightedWatershedsSegmentation(graph=gridGraph, seeds=seeds.ravel(),
                    edgeWeights=gridGraphEdgeStrength)
        seg = seg.reshape(shape)


    return seg

def distanceTransformWatersheds(pmap, preBinarizationMedianRadius=1, threshold = 0.5, preSeedSigma=0.75):
    """Superpixels for neuro data as in http://brainiac2.mit.edu/isbi_challenge/

    Use raw data and membrane probability maps to
    generate a over-segmentation suitable for neuro data

    Args:
        pmap (numpy.ndarray): Membrane probability in [0,1].
        preBinarizationMedianRadius (int) : Radius of
            diskMedian filter applied to the probability map
            before binarization. (default:{1})
        threshold (float) : threshold to binarize
            probability map  before applying
            the distance transform (default: {0.5})
        preSeedSigma (float) : smooth the distance
            transform image before getting the seeds.

    Raises:
        RuntimeError: if applied to data with wrong dimensionality
    """
    if pmap.ndim != 2:
        raise RuntimeError("Currently only implemented for 2D data")




    # pre-process pmap  / smooth pmap
    if preBinarizationMedianRadius >= 1 :
        toBinarize = filters.diskMedian(pmap, radius=preBinarizationMedianRadius)
        toBinarize -= toBinarize.min()
        toBinarize /= toBinarize.max()
    else:
        toBinarize = pmap


    # computing the distance transform inside and outside
    b1 = toBinarize < threshold

    #b0 = b1 == False

    dt1 = scipy.ndimage.morphology.distance_transform_edt(b1)
    #dt0 = scipy.ndimage.morphology.distance_transform_edt(b0)

    if preSeedSigma > 0.01:
        toSeedOn = filters.gaussianSmoothing(dt1, preSeedSigma)
    else:
        toSeedOn = dt1
    # find the seeds
    seeds  = localMaximaSeeds(toSeedOn)

    # compute  growing map
    a = filters.gaussianSmoothing(pmap, 0.75)
    b = filters.gaussianSmoothing(pmap, 3.00)
    c = filters.gaussianSmoothing(pmap, 9.00)
    d =  growMapB = numpy.exp(-0.2*dt1)
    growMap = (7.0*a + 4.0*b + 2.0*c + 1.0*d)/14.0


    # grow regions
    seg = seededWatersheds(growMap, seeds=seeds,
        method='edge_weighted',  acc='interpixel')
    return seg

def localMinima(image):
    """get the local minima of an image

    Get the local minima wrt a 4-neighborhood on an image.
    For a plateau, all pixels of this plateau are marked
    as minimum pixel.

    Args:
        image (numpy.ndarray): the input image

    Returns:
        (numpy.ndarray) : array which is 1 the minimum 0 elsewhere.

    """
    if image.ndim != 2:
        raise RuntimeError("localMinima is currently only implemented for 2D images")
    return localMaxima(-1.0*image)

def localMaxima(image):
    """get the local maxima of an image

    Get the local maxima wrt a 4-neighborhood on an image.
    For a plateau, all pixels of this plateau are marked
    as maximum pixel.

    Args:
        image (numpy.ndarray): the input image

    Returns:
        (numpy.ndarray) : array which is 1 the maximum 0 elsewhere.

    """
    if image.ndim != 2:
        raise RuntimeError("localMaxima is currently only implemented for 2D images")

    return  __peak_local_max(image, exclude_border=False, indices=False)

def connectedComponents(labels, dense=True, ignoreBackground=False):
    """get connected components of a label image

    Get connected components of an image w.r.t.
    a 4-neighborhood .
    This is a high level wrapper for
        :func:`nifty.graph.connectedComponentsFromNodeLabels`

    Args:
        labels (numpy.ndarray):
        dense (bool): should the return labeling be dense (default: {True})
        ignoreBackground (bool): should values of zero be excluded (default: {False})

    Returns:
        [description]
        [type]
    """
    shape = labels.shape
    gridGraph = graph.gridGraph(shape)

    ccLabels = graph.connectedComponentsFromNodeLabels(gridGraph,
            labels.ravel(), dense=dense,
            ignoreBackground=bool(ignoreBackground))

    return ccLabels.reshape(shape)

def localMinimaSeeds(image):
    """Get seed from local minima

    Get seeds by running connected components
    on the local minima.
    This is a high level wrapper around
    :func:`nifty.segmentation.localMinima`
    and :func:`nifty.segmentation.connectedComponents`

    Args:
        image: [description]

    Returns:
        [description]
        [type]

    Raises:
        RuntimeError: [description]
    """
    return localMaximaSeeds(-1.0 * image)

def localMaximaSeeds(image):
    """Get seed from local maxima

    Get seeds by running connected components
    on the local maxima.
    This is a high level wrapper around
    :func:`nifty.segmentation.localMinima`
    and :func:`nifty.segmentation.connectedComponents`

    Args:
        image: [description]

    Returns:
        [description]
        [type]

    Raises:
        RuntimeError: [description]
    """
    if image.ndim != 2:
        raise RuntimeError("localMaximaSeeds is currently only implemented for 2D images")

    lm = localMaxima(image)
    cc = connectedComponents(lm, dense=True, ignoreBackground=True)
    return cc

def markBoundaries(image, segmentation, color=None, thin=True):
    """Mark the boundaries in an image

    Mark boundaries in an image.

    Warning:

        The returned image shape is twice as large
        as the input if this is True.

    Args:
        image:  the input image
        segmentation:  the segmentation
        color (tuple) : the edge color(default: {(0,0,0)})
        thin (bool) : IF true, the image is interpolated and
            the boundaries are marked in the interpolated
            image. This will make the output twice as large.
    Returns:
        (numpy.ndarray) : image with marked boundaries. Note that
            result image has twice as large shape as the input if thin is True.
    """
    if color is None:
        color = (0,0,0)
    if thin:
        shape = segmentation.shape
        img2 =   __imresize(image, [2*s for s in shape])#, interp='nearest')
        #img2 = __zoom(segmentation.astype('float32'), 2, order=1)
        seg2 = __zoom(segmentation.astype('float32'), 2, order=0)
        seg2 = seg2.astype('uint32')

        return skimage.segmentation.mark_boundaries(img2, seg2.astype('uint32'), color=color)
    else:
        return skimage.segmentation.mark_boundaries(image, segmentation.astype('uint32'), color=color)

def segmentOverlay(image, segmentation, beta=0.5, zeroToZero=False, showBoundaries=True, color=None, thin=True):

    cmap = numpy.random.rand (int(segmentation.max()+1), 3)

    if zeroToZero:
        cmap[0,:] = 0
    cSeg = numpy.take(cmap, numpy.require(segmentation,dtype='int64'),axis=0)

    imgCp = image.astype('float32')
    if imgCp.ndim != 3:
        imgCp  = numpy.concatenate([imgCp[:,:,None]]*3,axis=2)


    mi = imgCp.min()
    ma = imgCp.max()

    if(ma-mi > 0.000001):
        imgCp -= mi
        imgCp /= (ma - mi)

    overlayImg =  (1.0-beta)*imgCp + (beta)*cSeg

    if showBoundaries:
        return markBoundaries(overlayImg, segmentation, color=color, thin=thin)
    else:
        return overlayImg





def randomColormap(size=10000, zeroToZero=False):
    cmap = numpy.random.rand (int(size),3)
    if zeroToZero:
        cmap[0,:] = 0
    cmap = __ListedColormap(cmap)
    return cmap



from __future__ import print_function
import unittest
import random
import numpy

import nifty
import nifty.graph
nlmc = nifty.graph.lifted_multicut


class TestLiftedGraphFeatures(unittest.TestCase):

    def generateGrid(self, gridSize):
        def nid(x, y):
            return x*gridSize[1] + y
        G = nifty.graph.UndirectedGraph
        g =  G(gridSize[0] * gridSize[1])
        for x in range(gridSize[0]):
            for y in range(gridSize[1]):

                u = nid(x,y)

                if x + 1 < gridSize[0]:

                    v = nid(x+1, y)
                    g.insertEdge(u, v)

                if y + 1 < gridSize[1]:

                    v = nid(x, y+1)
                    g.insertEdge(u, v)

        return g, nid

    def gridLiftedModel(self, gridSize = [3,2], bfsRadius=2, weightRange = [-1,1]):

        g,nid = self.generateGrid(gridSize)
        obj = nlmc.liftedMulticutObjective(g)
        graph = obj.graph
        liftedGraph = obj.liftedGraph


        # this should  add edges
        obj.insertLiftedEdgesBfs(bfsRadius)
        postEdges = liftedGraph.numberOfEdges


        for edge in liftedGraph.edges():
            u,v = liftedGraph.uv(edge)
            w = random.uniform(weightRange[0],weightRange[1])
            obj.setCost(u, v, w)

        return obj,nid


    def testLiftedUcFeatures(self):
        obj,nid = self.gridLiftedModel(gridSize=[100,100],bfsRadius=3)

        graph = obj.graph

        edgeIndicators = numpy.random.rand(graph.edgeIdUpperBound + 1) + 2.0
        edgeSizes      = numpy.random.rand(graph.edgeIdUpperBound + 1) + 0.5
        nodeSizes      = numpy.random.rand(graph.nodeIdUpperBound + 1) + 0.5

        features = nlmc.liftedUcmFeatures(
            objective=obj,
            edgeIndicators=edgeIndicators,
            edgeSizes=edgeSizes,
            nodeSizes=nodeSizes,
            sizeRegularizers=[0.01, 0.02, 0.3, 0.4]
        )

        featuresReg = features[::2,:]
        featuresRaw = features[1::2,:]

        self.assertEqual(featuresReg.shape[1], obj.numberOfLiftedEdges)
        self.assertEqual(featuresReg.shape[0], 4)
        self.assertEqual(featuresRaw.shape[0], 4)


        self.assertTrue(numpy.all(numpy.isfinite(featuresReg)))
        self.assertTrue(numpy.all(numpy.isfinite(featuresRaw)))

        self.assertGreaterEqual(featuresRaw.min(), 2.0)
        self.assertLessEqual(featuresRaw.max(), 3.0)

        self.assertGreaterEqual(featuresReg.min(), 0.0)


if __name__ == '__main__':
    unittest.main()



""" Multicut module of nifty

This module implements multicut
related functionality.

For more details, see seection :ref:`theory_multicut`.

"""

from __future__ import absolute_import
import sys
from functools import partial
from ._multicut import *
from .... import Configuration, LogLevel
from ... import (UndirectedGraph, EdgeContractionGraphUndirectedGraph)

__all__ = [
    "ilpSettings"
]
for key in _multicut.__dict__.keys():

    if key not in ["__spec__", "__doc__"]:
        try:
            value.__module__ = 'nifty.graph.opt.multicut'
        except Exception:
            continue
        __all__.append(key)


def ilpSettings(relativeGap=0.0, absoluteGap=0.0, memLimit=-1.0):
    """factory function to create :class:`IlpBackendSettigns` .

    factory function to create :class:`IlpBackendSettigns` .
    This settings might be consumed by ILP solvers
    as CPLEX GUROBI and GLPK.

    Args:
        relativeGap  (float): relative optimality gap (default: {0.0})
        absoluteGap  (float): absolute optimality gap (default: {0.0})
        memLimit (float): memory limit in mega-bites
            a value smaller as zero indicates no limit  (default: {-1.0})

    Returns:
        :class:`IlpBackendSettings`: ilpSettings
    """
    s = IlpBackendSettings()
    s.relativeGap = float(relativeGap)
    s.absoluteGap = float(absoluteGap)
    s.memLimit = float(memLimit)

    return s


def __extendMulticutObj(objectiveCls, objectiveName, graphCls):

    def getCls(prefix, postfix):
        return _multicut.__dict__[prefix+postfix]

    def getSettingsCls(baseName):
        S = getCls("__"+baseName + "SettingsType", objectiveName)
        return S

    def getMcCls(baseName):
        S = getCls(baseName, objectiveName)
        return S

    def getSettings(baseName):
        S = getSettingsCls(baseName)
        return S()

    def getSettingsAndFactoryCls(baseName):
        s = getSettings(baseName)
        F = getCls(baseName + "Factory", objectiveName)
        return s, F

    def factoryClsName(baseName):
        return baseName + "Factory" + objectiveName

    O = objectiveCls

    def verboseVisitor(visitNth=1, timeLimitSolver=float('inf'),
                       timeLimitTotal=float('inf'), logLevel=LogLevel.WARN):
        V = getMcCls("VerboseVisitor")
        return V(int(visitNth), float(timeLimitSolver), float(timeLimitTotal), logLevel)
    O.verboseVisitor = staticmethod(verboseVisitor)

    def loggingVisitor(visitNth=1, timeLimitSolver=float('inf'),
                       timeLimitTotal=float('inf'), logLevel=LogLevel.WARN):
        V = getMcCls("LoggingVisitor")
        return V(visitNth=int(visitNth),
                 timeLimitSolver=float(timeLimitSolver),
                 timeLimitTotal=float(timeLimitTotal),
                 logLevel=logLevel)
    O.loggingVisitor = staticmethod(loggingVisitor)

    def greedyAdditiveFactory(weightStopCond=0.0, nodeNumStopCond=-1.0, visitNth=1):
        s, F = getSettingsAndFactoryCls("MulticutGreedyAdditive")
        s.weightStopCond = float(weightStopCond)
        s.nodeNumStopCond = float(nodeNumStopCond)
        s.visitNth = int(visitNth)

        return F(s)
    O.greedyAdditiveFactory = staticmethod(greedyAdditiveFactory)
    O.greedyAdditiveFactory.__doc__ = """ create an instance of :class:`%s`

        Find approximate solutions via
        agglomerative clustering as in :cite:`beier_15_funsion`.

    Warning:
        This solver should be used to
        warm start other solvers with.
        This solver is very fast but
        yields rather suboptimal results.

    Args:
        weightStopCond (float): stop clustering when the highest
            weight in cluster-graph is lower as this value (default: {0.0})
        nodeNumStopCond (float): stop clustering when a cluster-graph
            reached a certain number of nodes.
            Numbers smaller 1 are interpreted as fraction
            of the graphs number of nodes.
            If nodeNumStopCond is smaller 0 this
            stopping condition is ignored  (default: {-1})
        visitNth (int) : only call the visitor each nth time.
            This is useful to avoid to many couts (default: {1}).
    Returns:
        %s : multicut factory
    """ % (factoryClsName("MulticutGreedyAdditive"), factoryClsName("MulticutGreedyAdditive"))

    def greedyFixationFactory(weightStopCond=0.0, nodeNumStopCond=-1.0, visitNth=1):
        s, F = getSettingsAndFactoryCls("MulticutGreedyFixation")
        s.weightStopCond = float(weightStopCond)
        s.nodeNumStopCond = float(nodeNumStopCond)
        s.visitNth = int(visitNth)

        return F(s)
    O.greedyFixationFactory = staticmethod(greedyFixationFactory)
    O.greedyFixationFactory.__doc__ = """ create an instance of :class:`%s`

    Warning:
        This solver should be used to
        warm start other solvers with.
        This solver is very fast but
        yields rather suboptimal results.

    Args:
        weightStopCond (float): stop clustering when the highest
            weight in cluster-graph is lower as this value (default: {0.0})
        nodeNumStopCond (float): stop clustering when a cluster-graph
            reached a certain number of nodes.
            Numbers smaller 1 are interpreted as fraction
            of the graphs number of nodes.
            If nodeNumStopCond is smaller 0 this
            stopping condition is ignored  (default: {-1})
        visitNth (int) : only call the visitor each nth time.
            This is useful to avoid to many couts (default: {1}).
    Returns:
        %s : multicut factory
    """ % (factoryClsName("MulticutGreedyFixation"), factoryClsName("MulticutGreedyFixation"))

    def warmStartGreedyDecorator(func):
        def func_wrapper(*args, **kwargs):
            warmStartGreedy = kwargs.pop('warmStartGreedy', False)
            warmStartKl = kwargs.pop("warmStartKl", False)
            greedyVisitNth = kwargs.pop('greedyVisitNth', 100)
            if warmStartGreedy:
                factories = [
                    greedyAdditiveFactory(visitNth=int(greedyVisitNth))
                ]
                if warmStartKl:
                    factories.append(kernighanLinFactory())
                factories.append(func(*args, **kwargs))
                return chainedSolversFactory(multicutFactories=factories)
            else:
                return func(*args, **kwargs)
        return func_wrapper

    def chainedSolversFactory(multicutFactories):
        s, F = getSettingsAndFactoryCls("ChainedSolvers")
        s. multicutFactories = multicutFactories
        return F(s)
    O.chainedSolversFactory = staticmethod(chainedSolversFactory)

    O.chainedSolversFactory.__doc__ = """ create an instance of :class:`%s`

        Chain multiple solvers
        such that each successor is warm-started with
        its predecessor solver.

    Warning:
        The solvers should be able to be warm started.

    Args:
        weightStopCond (float): stop clustering when the highest
            weight in cluster-graph is lower as this value (default: {0.0})
        nodeNumStopCond: stop clustering when a cluster-graph
            reached a certain number of nodes.
            Numbers smaller 1 are interpreted as fraction
            of the graphs number of nodes.
            If nodeNumStopCond is smaller 0 this
            stopping condition is ignored  (default: {-1})
    Returns:
        %s : multicut factory
    """ % (factoryClsName("ChainedSolvers"), factoryClsName("ChainedSolvers"))

    @warmStartGreedyDecorator
    def cgcFactory(doCutPhase=True, doGlueAndCutPhase=True, mincutFactory=None,
                   multicutFactory=None, doBetterCutPhase=False,
                   nodeNumStopCond=0.1, sizeRegularizer=1.0):
        if mincutFactory is None:
            if Configuration.WITH_QPBO:
                mincutFactory = graphCls.MincutObjective.mincutQpboFactory(improve=True)
            else:
                raise RuntimeError("default mincutFactory needs to be compiled WITH_QPBO")

        if Configuration.WITH_QPBO:
            s, F = getSettingsAndFactoryCls("Cgc")
            s.doCutPhase = bool(doCutPhase)
            s.doGlueAndCutPhase = bool(doGlueAndCutPhase)
            s.mincutFactory = mincutFactory
            if multicutFactory is not None:
                s.multicutFactory = multicutFactory
            s.doBetterCutPhase = bool(doBetterCutPhase)
            s.nodeNumStopCond = float(nodeNumStopCond)
            s.sizeRegularizer = float(sizeRegularizer)
            return F(s)
        else:
            raise RuntimeError("cgc need nifty to be compiled WITH_QPBO")
    O.cgcFactory = staticmethod(cgcFactory)
    O.cgcFactory.__module__ = "nifty.graph.opt.multicut"
    O.cgcFactory.__doc__ = """ create an instance of :class:`%s`

        Cut glue and cut as described in :cite:`beier_14_cut`.

    Warning:
        This solver should be warm started, otherwise
        the glue phase is very slow.
        Using :func:`greedyAdditiveFactory` to create
        a solver for warm starting is suggested.


    Note:
        In contrast to the OpenGM implementation we allow for
        arbitrary solvers to optimize the mincut problem.

    Args:
        doCutPhase: do recursive two coloring (default: {True})
        doGlueAndCutPhase: do re-opt of all pairs of clusters (default: {True})
        mincutFactory: mincutFactory for creating mincut solvers to solve subproblems (default: {None})
        multicutFactory: multicutFactory for creating multicut solvers to solve subproblems (default: {None})
        doBetterCutPhase: do a cut phase with multicut solvers instead of mincuts  (default: {False})
        nodeNumStopCond: If doBetterCutPhase is True, we use a agglomeration to
            create a set of clusters. Each cluster is then optimized with the solver from
            the multicutFactory. Values between 0 and 1 are interpreted as fraction
            of the total number of nodes in the graph (default: {0.1})
        sizeRegularizer: If doBetterCutPhase is True, we use a agglomeration to
            create a set of clusters.
            If this number is larger as zero, the clusters have about equal size (default: {1.0})
    Returns:
        %s : multicut factory
    """ % (factoryClsName("Cgc"), factoryClsName("Cgc"))

    def defaultMulticutFactory():
        return O.greedyAdditiveFactory()
    O.defaultMulticutFactory = staticmethod(defaultMulticutFactory)
    O.defaultFactory = staticmethod(defaultMulticutFactory)
    O.defaultFactory.__doc__ = """ create a instance of the default multicut solver factory.

        Currently the this function returns the same as
        :func:`greedyAdditiveFactory`

    Returns:
        %s : multicut factory
    """ % (factoryClsName("MulticutGreedyAdditive"))

    @warmStartGreedyDecorator
    def kernighanLinFactory(numberOfInnerIterations=sys.maxsize,
                            numberOfOuterIterations=100,
                            epsilon=1e-6):

        s, F = getSettingsAndFactoryCls("KernighanLin")
        s.numberOfInnerIterations = numberOfInnerIterations
        s.numberOfOuterIterations = numberOfOuterIterations
        s.epsilon = epsilon
        return F(s)
    O.kernighanLinFactory = staticmethod(kernighanLinFactory)
    O.kernighanLinFactory.__doc__ = """ create an instance of :class:`%s`

        Find approximate solutions via
        agglomerative clustering as in :cite:`TODO`.


    Args:
        numberOfInnerIterations (int): number of inner iterations (default: {sys.maxsize})
        numberOfOuterIterations (int): number of outer iterations        (default: {100})
        epsilon (float): epsilon   (default: { 1e-6})
        warmStartGreedy (bool): initialize with greedyAdditive  (default: {False})

    Returns:
        %s : multicut factory
    """ % tuple([factoryClsName("KernighanLin")]*2)

    def multicutDecomposerFactory(submodelFactory=None, fallthroughFactory=None, numberOfThreads=1):

        if submodelFactory is None:
            submodelFactory = MulticutObjectiveUndirectedGraph.defaultMulticutFactory()

        if fallthroughFactory is None:
            fallthroughFactory = O.defaultMulticutFactory()

        s, F = getSettingsAndFactoryCls("MulticutDecomposer")
        s.submodelFactory = submodelFactory
        s.fallthroughFactory = fallthroughFactory
        s.numberOfThreads = numberOfThreads
        return F(s)

    O.multicutDecomposerFactory = staticmethod(multicutDecomposerFactory)
    O.multicutDecomposerFactory.__doc__ = """ create an instance of :class:`%s`

        This solver tries to decompose the model into
        sub-models  as described in :cite:`alush_2013_simbad`.
        If a model decomposes into components such that there are no
        positive weighted edges between the components one can
        optimize each model separately.



    Note:
        Models might not decompose at all.

    Args:
        submodelFactory: multicut factory for solving subproblems
            if model decomposes (default: {:func:`defaultMulticutFactory()`})
        fallthroughFactory: multicut factory for solving subproblems
            if model does not decompose (default: {:func:`defaultMulticutFactory()`})

    Returns:
        %s : multicut factory
    """ % (factoryClsName("MulticutDecomposer"), factoryClsName("MulticutDecomposer"))

    def multicutIlpFactory(addThreeCyclesConstraints=True,
                           addOnlyViolatedThreeCyclesConstraints=True,
                           ilpSolverSettings=None,
                           ilpSolver=None):
        # default solver:
        if ilpSolver is None and Configuration.WITH_CPLEX:
            ilpSolver = 'cplex'
        if ilpSolver is None and Configuration.WITH_GUROBI:
            ilpSolver = 'gurobi'
        if ilpSolver is None and Configuration.WITH_GLPK:
            ilpSolver = 'glpk'
        if ilpSolver is None:
            raise RuntimeError("multicutIlpFactory needs either "
                               "'WITH_CPLEX', 'WITH_GUROBI'"
                               " or 'WITH_GLPK'  to be enabled")

        if ilpSolver == 'cplex':
            if not Configuration.WITH_CPLEX:
                raise RuntimeError("multicutIlpFactory with ilpSolver=`cplex` need nifty "
                    "to be compiled with WITH_CPLEX")
            s, F = getSettingsAndFactoryCls("MulticutIlpCplex")
        elif ilpSolver == 'gurobi':
            if not Configuration.WITH_GUROBI:
                raise RuntimeError("multicutIlpFactory with ilpSolver=`gurobi` need nifty "
                    "to be compiled with WITH_GUROBI")
            s, F = getSettingsAndFactoryCls("MulticutIlpGurobi")
        elif ilpSolver == 'glpk':
            if not Configuration.WITH_GLPK:
                raise RuntimeError("multicutIlpFactory with ilpSolver=`glpk` need nifty "
                    "to be compiled with WITH_GLPK")
            s, F = getSettingsAndFactoryCls("MulticutIlpGlpk")
        else:
            raise RuntimeError("%s is an unknown ilp solver" % str(ilpSolver))
        s.addThreeCyclesConstraints = bool(addThreeCyclesConstraints)
        s.addOnlyViolatedThreeCyclesConstraints = bool(addOnlyViolatedThreeCyclesConstraints)
        if ilpSolverSettings is None:
            ilpSolverSettings = ilpSettings()
        s.ilpSettings = ilpSolverSettings
        return F(s)

    O.multicutIlpFactory = staticmethod(multicutIlpFactory)
    if Configuration.WITH_CPLEX:
        O.multicutIlpCplexFactory = staticmethod(partial(multicutIlpFactory,ilpSolver='cplex'))
    if Configuration.WITH_GUROBI:
        O.multicutIlpGurobiFactory = staticmethod(partial(multicutIlpFactory,ilpSolver='gurobi'))
    if Configuration.WITH_GLPK:
        O.multicutIlpGlpkFactory = staticmethod(partial(multicutIlpFactory,ilpSolver='glpk'))

    O.multicutIlpFactory.__doc__ = """ create an instance of an ilp multicut solver.

        Find a global optimal solution by a cutting plane ILP solver
        as described in :cite:`Kappes-2011`
        and :cite:`andres_2011_probabilistic`


    Note:
        This might take very long for large models.

    Args:
        addThreeCyclesConstraints (bool) :
            explicitly add constraints for cycles
            of length 3 before opt (default: {True})
        addOnlyViolatedThreeCyclesConstraints (bool) :
            explicitly add all violated constraints for only violated cycles
            of length 3 before opt (default: {True})
        ilpSolverSettings (:class:`IlpBackendSettings`) :
            SettingsType of the ilp solver (default : {:func:`ilpSettings`})
        ilpSolver (str) : name of the solver. Must be in
            either "cplex", "gurobi" or "glpk".
            "glpk" is only capable of solving very small models.
            (default: {"cplex"}).

    Returns:
        %s or %s or %s : multicut factory for the corresponding solver

    """ % (
        factoryClsName("MulticutIlpCplex"),
        factoryClsName("MulticutIlpGurobi"),
        factoryClsName("MulticutIlpGlpk"),
    )

    # O.multicutIlpCplexFactory.__doc__ =
    # O.multicutIlpGurobiFactory
    # O.multicutIlpGlpkFactory

    if Configuration.WITH_LP_MP:
        def multicutMpFactory(
                mcFactory=None,
                numberOfIterations=1000,
                verbose=0,
                primalComputationInterval=100,
                standardReparametrization="anisotropic",
                roundingReparametrization="damped_uniform",
                tightenReparametrization="damped_uniform",
                tighten=True,
                tightenInterval=100,
                tightenIteration=10,
                tightenSlope=0.02,
                tightenConstraintsPercentage=0.1,
                minDualImprovement=0.,
                minDualImprovementInterval=0,
                timeout=0,
                numberOfThreads=1
        ):

            settings, factoryCls = getSettingsAndFactoryCls("MulticutMp")

            settings.mcFactory = mcFactory
            settings.numberOfIterations = numberOfIterations
            settings.verbose = verbose
            settings.primalComputationInterval = primalComputationInterval
            settings.standardReparametrization = standardReparametrization
            settings.roundingReparametrization = roundingReparametrization
            settings.tightenReparametrization = tightenReparametrization
            settings.tighten = tighten
            settings.tightenInterval = tightenInterval
            settings.tightenIteration = tightenIteration
            settings.tightenSlope = tightenSlope
            settings.tightenConstraintsPercentage = tightenConstraintsPercentage
            settings.minDualImprovement = minDualImprovement
            settings.minDualImprovementInterval = minDualImprovementInterval
            settings.timeout = timeout
            settings.numberOfThreads = numberOfThreads

            return factoryCls(settings)

        O.multicutMpFactory = staticmethod(multicutMpFactory)

    def fusionMoveSettings(mcFactory=None):
        if mcFactory is None:
            if Configuration.WITH_CPLEX:
                mcFactory = MulticutObjectiveUndirectedGraph.multicutIlpCplexFactory()
            else:
                mcFactory = MulticutObjectiveUndirectedGraph.defaultMulticutFactory()
        s = getSettings('FusionMove')
        s.mcFactory = mcFactory
        return s
    O.fusionMoveSettings = staticmethod(fusionMoveSettings)

    def watershedCcProposals(sigma=1.0, numberOfSeeds=0.1):
        s, F = getSettingsAndFactoryCls("WatershedProposalGenerator")
        s.sigma = float(sigma)
        s.numberOfSeeds = float(numberOfSeeds)
        return F(s)
    O.watershedCcProposals = staticmethod(watershedCcProposals)

    def interfaceFlipperCcProposals():
        s, F = getSettingsAndFactoryCls("InterfaceFlipperProposalGenerator")
        return F(s)
    O.interfaceFlipperCcProposals = staticmethod(interfaceFlipperCcProposals)

    def randomNodeColorCcProposals(numberOfColors=2):
        s, F = getSettingsAndFactoryCls("RandomNodeColorProposalGenerator")
        s.numberOfColors = int(numberOfColors)
        return F(s)
    O.randomNodeColorCcProposals = staticmethod(randomNodeColorCcProposals)

    @warmStartGreedyDecorator
    def ccFusionMoveBasedFactory(
        proposalGenerator=None,
        numberOfThreads=1, numberOfIterations=100,
        stopIfNoImprovement=10, fusionMove=None
    ):
        solverSettings, F = getSettingsAndFactoryCls("CcFusionMoveBased")

        if proposalGenerator is None:
            proposalGenerator = watershedCcProposals()
        if fusionMove is None:
            fusionMove = fusionMoveSettings()

        solverSettings.fusionMoveSettings = fusionMove
        solverSettings.proposalGenerator = proposalGenerator
        solverSettings.numberOfIterations = int(numberOfIterations)
        solverSettings.stopIfNoImprovement = int(stopIfNoImprovement)
        solverSettings.numberOfThreads = int(numberOfThreads)
        factory = F(solverSettings)
        return factory
    O.ccFusionMoveBasedFactory = staticmethod(ccFusionMoveBasedFactory)

    def perturbAndMapSettings(numberOfIterations=1000,
                              numberOfThreads=-1,
                              verbose=1,
                              noiseType='normal',
                              noiseMagnitude=1.0,
                              mcFactory=None):
        if mcFactory is None:
            mcFactory = fusionMoveBasedFactory(numberOfThreads=0)
        s = getSettings('PerturbAndMap')
        s.numberOfIterations = int(numberOfIterations)
        s.numberOfThreads = int(numberOfThreads)
        s.verbose = int(verbose)
        s.noiseMagnitude = float(noiseMagnitude)

        if(noiseType == 'normal'):
            s.noiseType = getMcCls('PerturbAndMap').NORMAL_NOISE
        elif(noiseType == 'uniform'):
            s.noiseType = getMcCls('PerturbAndMap').UNIFORM_NOISE
        elif(noiseType == 'makeLessCertain'):
            s.noiseType = graph.multicut.PerturbAndMapUndirectedGraph.MAKE_LESS_CERTAIN
        else:
            raise RuntimeError("'%s' is an unknown noise type. Must be 'normal' or 'uniform' or 'makeLessCertain' "%str(noiseType))

        s.mcFactory = mcFactory
        return s
    O.perturbAndMapSettings = staticmethod(perturbAndMapSettings)

    def perturbAndMap(objective, settings):
        pAndM = graph.multicut.perturbAndMap(objective, settings)
        return pAndM
    O.perturbAndMap = staticmethod(perturbAndMap)


__extendMulticutObj(MulticutObjectiveUndirectedGraph,
    "MulticutObjectiveUndirectedGraph",UndirectedGraph)
__extendMulticutObj(MulticutObjectiveEdgeContractionGraphUndirectedGraph,
    "MulticutObjectiveEdgeContractionGraphUndirectedGraph",EdgeContractionGraphUndirectedGraph)
del __extendMulticutObj



from __future__ import absolute_import
from ._opt import *
from .. import Configuration

# from . import multicut
# from . import lifted_multicut

# from multicut import *
# from lifted_multicut import *

__all__ = []

for key in _opt.__dict__.keys():
    __all__.append(key)



import os
import unittest
from shutil import rmtree

import numpy as np
import nifty
WITH_Z5 = nifty.Configuration.WITH_Z5
ON_WIN = os.name == 'nt'

try:
    import z5py
    WITH_Z5PY = True
except ImportError:
    WITH_Z5PY = False


@unittest.skipIf(ON_WIN, "Fails on windows")
@unittest.skipUnless(WITH_Z5 and WITH_Z5PY, "Need z5 support")
class TestSkeletons(unittest.TestCase):

    def setUp(self):
        if not os.path.exists('./tmp'):
            os.mkdir('./tmp')

        # make segmenation data
        seg = np.zeros((100, 100, 100), dtype='uint64')
        seg[:50] = 1
        seg[50:, 50:, 50:] = 2
        seg[:50, :50, :50] = 3
        f_seg = z5py.File('./tmp/seg.n5', use_zarr_format=False)
        ds = f_seg.create_dataset('seg', shape=seg.shape, chunks=(20, 20, 20), dtype=seg.dtype)
        ds[:] = seg

        # make skeletons
        f_skels = z5py.File('./tmp/skels.n5', use_zarr_format=False)
        g = f_skels.require_group('skels')
        # skeleton 1: only in label 3
        skel1 = np.array([[0, 0, 0, 0],
                          [1, 10, 10, 10],
                          [2, 15, 15, 15],
                          [3, 20, 20, 20],
                          [4, 25, 25, 25],
                          [5, 35, 35, 35]], dtype='uint64')
        g1 = g.create_group('1')
        c1 = g1.create_dataset('coordinates', shape=skel1.shape,
                               chunks=skel1.shape, dtype='uint64')
        c1[:] = skel1
        # skeleton 2: in label 0 and 2
        skel2 = np.array([[0, 60, 0, 0],
                          [1, 70, 10, 10],
                          [2, 75, 25, 25],
                          [3, 75, 45, 45],
                          [4, 80, 55, 55],
                          [5, 85, 65, 65]], dtype='uint64')
        g2 = g.create_group('2')
        c2 = g2.create_dataset('coordinates', shape=skel2.shape,
                               chunks=skel2.shape, dtype='uint64')
        c2[:] = skel2

    def tearDown(self):
        if os.path.exists('./tmp'):
            rmtree('./tmp')

    def test_nodes(self):
        import nifty.skeletons as nskel
        metrics = nskel.SkeletonMetrics('./tmp/seg.n5', 'seg', './tmp/skels.n5', 'skels', [1, 2], 1)
        out = metrics.getNodeAssignments()
        self.assertEqual(list(out.keys()), [1, 2])
        out1 = [out[1][k] for k in sorted(out[1].keys())]
        out2 = [out[2][k] for k in sorted(out[2].keys())]
        self.assertEqual(list(out1), 6 * [3])
        self.assertEqual(list(out2), [0, 0, 0, 0, 2, 2])


if __name__ == '__main__':
    unittest.main()



import unittest
import random
import numpy

import nifty
import nifty.graph
import nifty.graph.opt.minstcut


class TestMinstcutObjective(unittest.TestCase):
    def setUp(self):
        numpy.random.seed(7)

    def generateGrid(self, gridSize):
        def nid(x, y):
            return x*gridSize[1] + y
        G = nifty.graph.UndirectedGraph
        g =  G(gridSize[0] * gridSize[1])
        for x in range(gridSize[0]):
            for y in range(gridSize[1]):

                u = nid(x,y)

                if x + 1 < gridSize[0]:

                    v = nid(x+1, y)
                    g.insertEdge(u, v)

                if y + 1 < gridSize[1]:

                    v = nid(x, y+1)
                    g.insertEdge(u, v)
        return g, nid

    def testGridModel(self):
        beta = .7
        gridSize = [5,5]
        weightRange = [0,1]

        g,nid = self.generateGrid(gridSize)
        d = weightRange[1] - weightRange[0]

        w = numpy.zeros(g.numberOfEdges)
        u = numpy.zeros((g.numberOfNodes,2))

        labels = numpy.zeros(g.nodeIdUpperBound+1,dtype='uint8')

        for x in range(gridSize[0]):
            for y in range(gridSize[1]):

                nodeId = nid(x,y)
                if x in range(1,gridSize[0] - 1):
                    if y in range(1,gridSize[1] - 1):
                        u[nodeId,0] = .1
                        u[nodeId,1] = .9
                        labels[nodeId] = 1
                    else:
                        u[nodeId,0] = .8
                        u[nodeId,1] = .2
                        labels[nodeId] = 0
                else:
                    u[nodeId,0] = .8
                    u[nodeId,1] = .2
                    labels[nodeId] = 0

        for x in range(gridSize[0]):
            for y in range(gridSize[1]):

                p = nid(x,y)

                if x + 1 < gridSize[0]:
                    q = nid(x+1,y)
                    weightId = g.findEdge(p,q)
                    if (labels[p] == labels[q]):
                        w[weightId] = 0
                    else:
                        w[weightId] = beta
                if y + 1 < gridSize[1]:
                    q = nid(x, y + 1)
                    weightId = g.findEdge(p,q)
                    if (labels[p] == labels[q]):
                        w[weightId] = 0
                    else:
                        w[weightId] = beta

        obj = nifty.graph.opt.minstcut.minstcutObjective(g,w,u)
        value = obj.evalNodeLabels(labels)
        self.assertAlmostEqual(value,29.3)
        return obj


if __name__ == '__main__':
    unittest.main()



import unittest
import random
import numpy
import nifty
import nifty.graph


class TestLiftedMulticutSolver(unittest.TestCase):
    def setUp(self):
        numpy.random.seed(7)

    def generateGrid(self, gridSize):
        def nid(x, y):
            return x*gridSize[1] + y
        G = nifty.graph.UndirectedGraph
        g =  G(gridSize[0] * gridSize[1])
        for x in range(gridSize[0]):
            for y in range(gridSize[1]):

                u = nid(x,y)

                if x + 1 < gridSize[0]:

                    v = nid(x+1, y)
                    g.insertEdge(u, v)

                if y + 1 < gridSize[1]:

                    v = nid(x, y+1)
                    g.insertEdge(u, v)
        return g, nid

    def gridModel(self, gridSize = [5,5],  weightRange = [-2,1]):
        g,nid = self.generateGrid(gridSize)
        d = weightRange[1] - weightRange[0]
        w = numpy.random.rand(g.numberOfEdges)*d + float(weightRange[0])
        # print(w.min(),w.max())
        obj = nifty.graph.multicut.multicutObjective(g,w)
        return obj

    def _testGridModelImpl(self, factory, gridSize=[6,6]):

        objective = self.gridModel(gridSize=gridSize)

        # with verbose visitor
        solver = factory.create(objective)
        visitor = objective.verboseVisitor(1000)

        self.assertEqual(visitor.timeLimitTotal, float('inf'))
        self.assertEqual(visitor.timeLimitSolver, float('inf'))

        arg = solver.optimize(visitor)

        self.assertEqual(visitor.timeLimitTotal, float('inf'))
        self.assertEqual(visitor.timeLimitSolver, float('inf'))

        # with verbose visitor
        solver = factory.create(objective)
        visitor = objective.verboseVisitor(1)

        self.assertEqual(visitor.timeLimitTotal, float('inf'))
        self.assertEqual(visitor.timeLimitSolver, float('inf'))

        arg = solver.optimize(visitor)

        self.assertEqual(visitor.timeLimitTotal, float('inf'))
        self.assertEqual(visitor.timeLimitSolver, float('inf'))

        # without any visitor
        solver = factory.create(objective)
        arg = solver.optimize()

    @unittest.skipUnless(nifty.Configuration.WITH_QPBO, "need qpbo")
    def testCgc(self):
        objective = self.gridModel(gridSize=[6,6])
        solver = objective.cgcFactory(True,True).create(objective)
        visitor = objective.verboseVisitor(1)
        arg = solver.optimize(visitor)

    def testGreedyAdditive(self):
        Obj = nifty.graph.UndirectedGraph.MulticutObjective
        self._testGridModelImpl(Obj.greedyAdditiveFactory(), gridSize=[6,6])

    def testDefault(self):
        Obj = nifty.graph.UndirectedGraph.MulticutObjective
        self._testGridModelImpl(Obj.defaultFactory(), gridSize=[6,6])

    def testMulticutDecomposer(self):
        Obj = nifty.graph.UndirectedGraph.MulticutObjective
        self._testGridModelImpl(Obj.multicutDecomposerFactory(), gridSize=[6,6])

    def testChainedSolvers(self):
        Obj = nifty.graph.UndirectedGraph.MulticutObjective
        a = Obj.greedyAdditiveFactory()
        b = Obj.defaultFactory()
        c = Obj.multicutDecomposerFactory()
        self._testGridModelImpl(Obj.chainedSolversFactory([a,b,c]), gridSize=[6,6])

    def testCcFusionMoveBasedFactory(self):
        Obj = nifty.graph.UndirectedGraph.MulticutObjective
        self._testGridModelImpl(Obj.ccFusionMoveBasedFactory(), gridSize=[10,10])

        self._testGridModelImpl(
            Obj.ccFusionMoveBasedFactory(proposalGenerator= Obj.watershedCcProposals()),
            gridSize=[10,10])

        self._testGridModelImpl(
            Obj.ccFusionMoveBasedFactory(proposalGenerator= Obj.interfaceFlipperCcProposals()),
            gridSize=[10,10])

        self._testGridModelImpl(
            Obj.ccFusionMoveBasedFactory(proposalGenerator= Obj.randomNodeColorCcProposals()),
            gridSize=[10,10])

    @unittest.skipUnless(nifty.Configuration.WITH_CPLEX, "need cplex")
    def testMulticutIlpCplex(self):
        Obj = nifty.graph.UndirectedGraph.MulticutObjective
        self._testGridModelImpl(Obj.multicutIlpCplexFactory(), gridSize=[5,5])

    @unittest.skipUnless(nifty.Configuration.WITH_GUROBI, "need gurobi")
    def testMulticutIlpGurobi(self):
        Obj = nifty.graph.UndirectedGraph.MulticutObjective
        self._testGridModelImpl(Obj.multicutIlpGurobiFactory(), gridSize=[5,5])

    @unittest.skipUnless(nifty.Configuration.WITH_GLPK, "need glpk")
    def testMulticutIlpGlpk(self):
        Obj = nifty.graph.UndirectedGraph.MulticutObjective
        objective = self.gridModel(gridSize=[4,5])
        factory = Obj.multicutIlpGlpkFactory()
        solver = factory.create(objective)
        visitor = objective.verboseVisitor(1000)
        self.assertEqual(visitor.timeLimitTotal, float('inf'))
        self.assertEqual(visitor.timeLimitSolver, float('inf'))
        arg = solver.optimize(visitor)


if __name__ == '__main__':
    unittest.main()



from __future__ import print_function
import unittest
import os
import numpy as np

import nifty
import nifty.graph.rag as nrag
WITH_HDF5 = nifty.Configuration.WITH_HDF5

try:
    import h5py
    WITH_H5PY = True
except ImportError:
    WITH_H5PY = False


class TestHdf5Serialization(unittest.TestCase):

    def setUp(self):
        self.rag_path = './rag_tmp.h5'
        self.seg_path = './seg_tmp.h5'

    def tearDown(self):
        if os.path.exists(self.rag_path):
            os.remove(self.rag_path)
        if os.path.exists(self.seg_path):
            os.remove(self.seg_path)

    @staticmethod
    def make_random_stacked_seg(shape):
        seg = np.zeros(shape, dtype = 'uint32')
        label = 0
        for z in range(shape[0]):
            if z > 0:
                label = np.max(seg[z-1]) + 1
            for y in range(shape[1]):
                for x in range(shape[2]):
                    seg[z, y, x] = label
                    if np.random.randint(0,10) > 8:
                        label += 1
        return seg

    @staticmethod
    def check_stacked(seg):
        prev_max = -1
        for z in range(seg.shape[0]):
            min_label = seg[z].min()
            max_label = seg[z].max()
            assert min_label == prev_max + 1
            prev_max = max_label

    @unittest.skipUnless(WITH_HDF5 and WITH_H5PY, "need nifty hdf5 and h5py")
    def test_hdf5_serialization(self):
        import nifty.hdf5 as nh5
        seg = self.make_random_stacked_seg((20, 100, 100))
        self.check_stacked(seg)
        n_labels = seg.max() + 1

        with h5py.File(self.seg_path, 'a') as f:
            f.create_dataset('data', data=seg)

        label_f = nh5.openFile(self.seg_path)
        labels = nh5.Hdf5ArrayUInt32(label_f, 'data')

        rag = nrag.gridRagStacked2DHdf5(labels, n_labels, -1)

        nifty.graph.rag.writeStackedRagToHdf5(rag, self.rag_path)
        rag_read = nrag.readStackedRagFromHdf5(labels, n_labels, self.rag_path)

        self.assertTrue((rag.uvIds() == rag_read.uvIds()).all())
        self.assertTrue((rag.numberOfEdges == rag_read.numberOfEdges))
        self.assertTrue((rag.numberOfNodes == rag_read.numberOfNodes))
        self.assertTrue((rag.minMaxLabelPerSlice() == rag_read.minMaxLabelPerSlice()).all())
        self.assertTrue((rag.numberOfNodesPerSlice() == rag_read.numberOfNodesPerSlice()).all())
        self.assertTrue((rag.numberOfInSliceEdges() == rag_read.numberOfInSliceEdges()).all())
        self.assertTrue((rag.numberOfInBetweenSliceEdges() == rag_read.numberOfInBetweenSliceEdges()).all())
        self.assertTrue((rag.inSliceEdgeOffset() == rag_read.inSliceEdgeOffset()).all())
        self.assertTrue((rag.betweenSliceEdgeOffset() == rag_read.betweenSliceEdgeOffset()).all())
        self.assertTrue((rag.totalNumberOfInSliceEdges == rag_read.totalNumberOfInSliceEdges))
        self.assertTrue((rag.totalNumberOfInBetweenSliceEdges == rag_read.totalNumberOfInBetweenSliceEdges))


if __name__ == '__main__':
    unittest.main()



import unittest
import nifty.tools as nt


class TestBlocking(unittest.TestCase):

    def testBlockingBlocksPerAxis(self):

        roiBegin =   [0,0,0]
        roiEnd =     [5,6,7]
        blockShape = [3,3,3]
        blockShift = [0,0,0]

        blocking = nt.blocking(roiBegin=roiBegin,roiEnd=roiEnd,
                                 blockShape=blockShape, blockShift=blockShift)

        blocksPerAxis = blocking.blocksPerAxis

        self.assertEqual(blocksPerAxis[0],2)
        self.assertEqual(blocksPerAxis[1],2)
        self.assertEqual(blocksPerAxis[2],3)

        roiBegin =   [0,0,0]
        roiEnd =     [5,6,7]
        blockShape = [3,3,3]
        blockShift = [0,1,0]

        blocking = nt.blocking(roiBegin=roiBegin,roiEnd=roiEnd,
                                 blockShape=blockShape, blockShift=blockShift)


        blocksPerAxis = blocking.blocksPerAxis

        self.assertEqual(blocksPerAxis[0],2)
        self.assertEqual(blocksPerAxis[1],3)
        self.assertEqual(blocksPerAxis[2],3)

    def testBlocking2d(self):

        roiBegin =   [0,0]
        roiEnd =     [5,7]
        blockShape = [3,3]
        blockShift = [0,0]

        blocking = nt.blocking(roiBegin=roiBegin,roiEnd=roiEnd,
                                 blockShape=blockShape, blockShift=blockShift)

        blocksPerAxis = blocking.blocksPerAxis

        self.assertEqual(blocksPerAxis[0],2)
        self.assertEqual(blocksPerAxis[1],3)

        halo = [2,2]

        blocks = [blocking.getBlock(i) for i in range(blocking.numberOfBlocks)]
        blocksWithHalo = [blocking.getBlockWithHalo(i, halo)
                          for i in range(blocking.numberOfBlocks)]

        self.assertEqual(blocks[0].begin,[0,0])
        self.assertEqual(blocks[0].end,  [3,3])

        self.assertEqual(blocksWithHalo[0].innerBlock.begin,[0,0])
        self.assertEqual(blocksWithHalo[0].innerBlock.end,  [3,3])
        self.assertEqual(blocksWithHalo[0].outerBlock.begin,[0,0])
        self.assertEqual(blocksWithHalo[0].outerBlock.end,  [5,5])

        self.assertEqual(blocks[1].begin,[0,3])
        self.assertEqual(blocks[1].end,  [3,6])
        self.assertEqual(blocksWithHalo[1].innerBlock.begin,[0,3])
        self.assertEqual(blocksWithHalo[1].innerBlock.end,  [3,6])
        self.assertEqual(blocksWithHalo[1].outerBlock.begin,[0,1])
        self.assertEqual(blocksWithHalo[1].outerBlock.end,  [5,7])

        self.assertEqual(blocks[2].begin,[0,6])
        self.assertEqual(blocks[2].end,  [3,7])
        self.assertEqual(blocksWithHalo[2].innerBlock.begin,[0,6])
        self.assertEqual(blocksWithHalo[2].innerBlock.end,  [3,7])
        self.assertEqual(blocksWithHalo[2].outerBlock.begin,[0,4])
        self.assertEqual(blocksWithHalo[2].outerBlock.end,  [5,7])

        self.assertEqual(blocks[3].begin,[3,0])
        self.assertEqual(blocks[3].end,  [5,3])
        self.assertEqual(blocksWithHalo[3].innerBlock.begin,[3,0])
        self.assertEqual(blocksWithHalo[3].innerBlock.end,  [5,3])
        self.assertEqual(blocksWithHalo[3].outerBlock.begin,[1,0])
        self.assertEqual(blocksWithHalo[3].outerBlock.end,  [5,5])

        self.assertEqual(blocks[4].begin,[3,3])
        self.assertEqual(blocks[4].end,  [5,6])
        self.assertEqual(blocksWithHalo[4].innerBlock.begin,[3,3])
        self.assertEqual(blocksWithHalo[4].innerBlock.end,  [5,6])
        self.assertEqual(blocksWithHalo[4].outerBlock.begin,[1,1])
        self.assertEqual(blocksWithHalo[4].outerBlock.end,  [5,7])

        self.assertEqual(blocks[5].begin,[3,6])
        self.assertEqual(blocks[5].end,  [5,7])
        self.assertEqual(blocksWithHalo[5].innerBlock.begin,[3,6])
        self.assertEqual(blocksWithHalo[5].innerBlock.end,  [5,7])
        self.assertEqual(blocksWithHalo[5].outerBlock.begin,[1,4])
        self.assertEqual(blocksWithHalo[5].outerBlock.end,  [5,7])

    def testNeighbors(self):
        blocking = nt.blocking(roiBegin=[0, 0, 0],
                               roiEnd=[40, 35, 17],
                               blockShape=[5, 5, 5])
        neighbors = {}
        for block_id in range(blocking.numberOfBlocks):
            nbrs = [blocking.getNeighborId(block_id,
                                           axis=i // 2,
                                           lower=bool(i % 2)) for i in range(6)]
            neighbors[block_id] = [nbr for nbr in nbrs if nbr != -1]

        for block_id in range(blocking.numberOfBlocks):
            nbrs = neighbors[block_id]
            for nbr_id in nbrs:
                self.assertTrue(block_id in neighbors[nbr_id])

    def testNeighborsToyExample(self):
        blocking = nt.blocking(roiBegin=[0, 0],
                               roiEnd=[10, 10],
                               blockShape=[5, 5])

        neighbors = {}
        for b in range(blocking.numberOfBlocks):
            nbrs = [blocking.getNeighborId(b, i // 2, bool(i % 2)) for i in range(4)]
            nbrs = [nbr for nbr in nbrs if nbr != -1]
            self.assertEqual(len(nbrs), 2)
            nbrs.sort()
            neighbors[b] = nbrs

        self.assertEqual(neighbors[0][0], 1)
        self.assertEqual(neighbors[0][1], 2)

        self.assertEqual(neighbors[1][0], 0)
        self.assertEqual(neighbors[1][1], 3)

        self.assertEqual(neighbors[2][0], 0)
        self.assertEqual(neighbors[2][1], 3)

        self.assertEqual(neighbors[3][0], 1)
        self.assertEqual(neighbors[3][1], 2)


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import
import numpy

from .. import Configuration

from ._z5 import *

__all__ = []

if Configuration.WITH_Z5:
    from ._z5 import *
    from . import _z5 as __z5
    for key in __z5.__dict__.keys():
        __all__.append(key)
        try:
            __z5.__dict__[key].__module__ = 'nifty.z5'
        except:
            pass
else:
    pass


if Configuration.WITH_Z5:

    # convenience wrapper around z5 dataset wrapper
    def datasetWrapper(dtype, *args, **kwargs):

        if numpy.dtype(dtype) == numpy.dtype("uint8"):
            return DatasetWrapperUint8(*args, **kwargs)
        elif numpy.dtype(dtype) == numpy.dtype("uint16"):
            return DatasetWrapperUint16(*args, **kwargs)
        elif numpy.dtype(dtype) == numpy.dtype("uint32"):
            return DatasetWrapperUint32(*args, **kwargs)
        elif numpy.dtype(dtype) == numpy.dtype("uint64"):
            return DatasetWrapperUint64(*args, **kwargs)

        elif numpy.dtype(dtype) == numpy.dtype("int8"):
            return DatasetWrapperInt8(*args, **kwargs)
        elif numpy.dtype(dtype) == numpy.dtype("int16"):
            return DatasetWrapperInt16(*args, **kwargs)
        elif numpy.dtype(dtype) == numpy.dtype("int32"):
            return DatasetWrapperInt32(*args, **kwargs)
        elif numpy.dtype(dtype) == numpy.dtype("int64"):
            return DatasetWrapperInt64(*args, **kwargs)

        elif numpy.dtype(dtype) == numpy.dtype("float32"):
            return DatasetWrapperFloat32(*args, **kwargs)
        elif numpy.dtype(dtype) == numpy.dtype("float64"):
            return DatasetWrapperFloat64(*args, **kwargs)

        else:
            raise RuntimeError("Datatype %s not supported!" % (str(dtype),))


    # TODO infer datatype from the dataset
    def nearestUpsampling(dtype, inPath, samplingFactor, outPath, numberOfThreads=-1):
        if numpy.dtype(dtype) == numpy.dtype("uint8"):
            nearestUpsamplingUint8(inPath, samplingFactor, outPath, numberOfThreads)
        else:
            raise RuntimeError("Datatype %s not supported!" % (str(dtype),))



from __future__ import absolute_import
from __future__ import print_function

from ._nifty import *

import types
import numpy
import time
import sys

from . import graph
from . import tools
from . import ufd


class Timer:
    def __init__(self, name=None, verbose=True):
        """Timer class as with statement

        Time pieces of code with a with statement timer

        Examples:

            import nifty
            with nifty.Timer() as t:
                import time
                time.sleep()

        Args:
            name: name to print (default: {None})
            verbose: do printout (default: {True})
        """
        self.name = name
        self.verbose = verbose

    def __enter__(self):
        if self.verbose:
            if self.name is not None:
                print(self.name,"...")


        self.start = time.clock()
        return self

    def __exit__(self, *args):
        self.end = time.clock()
        self.elapsedTime = self.end - self.start
        if self.verbose:
            if self.name is not None:
                print(self.name, "took", self.elapsedTime, "sec")
            else:
                print("took", self.elapsedTime, "sec")



from __future__ import print_function
import nifty
import numpy



G = nifty.graph.UndirectedGraph
CG = G.EdgeContractionGraph

GMCO = G.MulticutObjective
CGMCO = CG.MulticutObjective



def testEdgeContractionGraph():

    g =  nifty.graph.UndirectedGraph(4)
    edges =  numpy.array([[0,1],[0,2],[0,3]],dtype='uint64')
    g.insertEdges(edges)





    class MyCb(nifty.graph.EdgeContractionGraphCallback):
        def __init__(self):
            super(MyCb, self).__init__()

            self.nCallsContractEdge = 0
            self.nCallsMergeEdges = 0
            self.nCallsMergeNodes = 0
            self.nCallscontractEdgeDone = 0

        def contractEdge(self, edge):
            self.nCallsContractEdge += 1

        def mergeEdges(self, alive, dead):
            self.nCallsMergeEdges += 1

        def mergeNodes(self, alive, dead):
            self.nCallsMergeNodes += 1

        def contractEdgeDone(self, edge):
            self.nCallscontractEdgeDone += 1



    cb = MyCb()
    ecg = nifty.graph.edgeContractionGraph(g,cb)
    ecg.contractEdge(0)

    assert cb.nCallsContractEdge == 1
    assert cb.nCallsMergeNodes == 1
    assert cb.nCallscontractEdgeDone == 1




from __future__ import print_function
import nifty
import numpy

import numpy
import vigra
import glob
import os
from functools import partial


nrag = nifty.graph.rag
ngala = nifty.graph.gala
ngraph = nifty.graph
G = nifty.graph.UndirectedGraph

def make_dataset(numberOfImages = 10, noise=1.0,shape=(100,100)):

    numpy.random.seed(42)
    imgs = []
    gts = []


    for i in range(numberOfImages):

        gtImg = numpy.zeros(shape)
        gtImg[0:shape[0]/2,:] = 1

        gtImg[shape[0]/4: 3*shape[0]/4, shape[0]/4: 3*shape[0]/4]  = 2

        ra = numpy.random.randint(180)
        #print ra 

        gtImg = vigra.sampling.rotateImageDegree(gtImg.astype(numpy.float32),int(ra),splineOrder=0)
        grad = vigra.filters.gaussianGradientMagnitude(gtImg.astype('float32'),1.0)


        for x in range(6):
            ra2 = numpy.random.randint(low=0,high=2,size=gtImg.size)
            ra2 = ra2.reshape(gtImg.shape)
            #print ra 
            grad*=ra2

        grad = grad.squeeze()

        img = numpy.random.random(shape)*float(0.00001)
        grad+=img
        #if i<1 :
        #    vigra.imshow(img)
        #    vigra.show()

        imgs.append(grad.astype('float32'))
        gts.append(gtImg)


    return imgs,gts

def makeRag(raw, showSeg = False):
    #raw = vigra.gaussianSmoothing(raw,1.0)
    ew = vigra.filters.hessianOfGaussianEigenvalues(-1.0*raw, 2.3)[:,:,0]
    seg, nseg = vigra.analysis.watershedsNew(ew)

    #seg, nseg = vigra.analysis.slicSuperpixels(raw,intensityScaling=4.5, seedDistance=20)

    seg = seg.squeeze()
    if showSeg:
        vigra.segShow(raw, seg)
        vigra.show()

    # get the rag
    seg -= 1
    assert seg.min() == 0
    assert seg.max() == nseg -1 
    return nifty.graph.rag.gridRag(seg)

def makeFeatureOpFromChannel(rag, data, minVal=None, maxVal=None):

    #  feature accumulators
    if minVal is None:
        minVal = float(data.min())

    if maxVal is None:
        maxVal = float(data.max())

    edgeFeatures = nifty.graph.rag.defaultAccEdgeMap(rag, minVal, maxVal)
    nodeFeatures = nifty.graph.rag.defaultAccNodeMap(rag, minVal, maxVal)

    # accumulate features
    nrag.gridRagAccumulateFeatures(graph=rag,data=data,
        edgeMap=edgeFeatures, nodeMap=nodeFeatures)


    fOp = ngala.galaDefaultAccFeature(graph=rag, edgeFeatures=edgeFeatures, nodeFeatures=nodeFeatures)
    
    return fOp,minVal, maxVal

def makeFeatureOp(rag, raw, minVals=None, maxVals=None):


    filterFuc = [
        partial(vigra.filters.gaussianSmoothing,sigma=0.5),
        partial(vigra.filters.gaussianSmoothing,sigma=1.0),
        partial(vigra.filters.gaussianSmoothing,sigma=2.0),
        partial(vigra.filters.gaussianSmoothing,sigma=4.0),
        partial(vigra.filters.gaussianGradientMagnitude,sigma=1.0),
        partial(vigra.filters.gaussianGradientMagnitude,sigma=2.0),
        partial(vigra.filters.gaussianGradientMagnitude,sigma=4.0),
        partial(vigra.filters.hessianOfGaussianEigenvalues,scale=1.0),
        partial(vigra.filters.hessianOfGaussianEigenvalues,scale=2.0),
        partial(vigra.filters.hessianOfGaussianEigenvalues,scale=4.0),
        partial(vigra.filters.structureTensorEigenvalues,innerScale=1.0,outerScale=2.0),
        partial(vigra.filters.structureTensorEigenvalues,innerScale=2.0,outerScale=4.0),
        partial(vigra.filters.structureTensorEigenvalues,innerScale=4.0,outerScale=8.0)
    ]
    fCollection = ngala.galaFeatureCollection(rag)

    minVals_ = []
    maxVals_ = []
    c = 0
    for f in filterFuc:
        res = f(raw).squeeze()
        if res.ndim == 2:
            res = res[:, :, None]
        for c in range(res.shape[2]):
            resC = res[:, :, c]

            if minVals is not None:
                minv = minVals[c]
            else:
                minv = resC.min()
            if maxVals is not None:
                maxv = maxVals[c]
            else:
                maxv = resC.max()
            minVals_.append(minv)
            maxVals_.append(maxv)
            op, _minVal , _maxVal =  makeFeatureOpFromChannel(rag, resC, minVal=minv, maxVal=maxv)
            fCollection.addFeatures(op)
            c +=1

    return fCollection, minVals_, maxVals_

def makeEdgeGt(rag, gt):
    # get the gt
    nodeGt = nrag.gridRagAccumulateLabels(rag, gt)
    uvIds = rag.uvIds()
    edgeGt = (nodeGt[uvIds[:,0]] != nodeGt[uvIds[:,1]]).astype('double')
    return edgeGt

def test_mcgala():



    # get the dataset
    imgs,gts = make_dataset(10, noise=4.5, shape=(200,200))

    Obj = G.MulticutObjective
    CG = G.EdgeContractionGraph
    CGObj = CG.MulticutObjective


    greedyFactory = Obj.greedyAdditiveFactory()
    ilpFactory = Obj.multicutIlpFactory(ilpSolver='cplex',
        addThreeCyclesConstraints=False,
        addOnlyViolatedThreeCyclesConstraints=False
        #memLimit= 0.01
    )
    fmFactoryA = CGObj.fusionMoveBasedFactory(
        #fusionMove=CGObj.fusionMoveSettings(mcFactory=greedyFactory),
        fusionMove=CGObj.fusionMoveSettings(mcFactory=ilpFactory),
        #proposalGen=nifty.greedyAdditiveProposals(sigma=30,nodeNumStopCond=-1,weightStopCond=0.0),
        proposalGen=CGObj.watershedProposals(sigma=1,seedFraction=0.1),
        numberOfIterations=10,
        numberOfParallelProposals=40, # no effect if nThreads equals 0 or 1
        numberOfThreads=40,
        stopIfNoImprovement=2,
        fuseN=2,
    )




    ragTrain  = makeRag(imgs[0], showSeg= True)
    fOpTrain, minVal, maxVal = makeFeatureOp(ragTrain, imgs[0])
    edgeGt = makeEdgeGt(ragTrain, gts[0])


    cOrderSettigns = G.galaContractionOrderSettings(
        mcMapFactory=fmFactoryA,
        runMcMapEachNthTime=10)

    # gala class
    settings = G.galaSettings(threshold0=0.1, threshold1=0.9, thresholdU=0.1,
                              numberOfEpochs=1, numberOfTrees=200,
                              contractionOrderSettings=cOrderSettigns
                              #mapFactory=fmFactoryA,
                              #perturbAndMapFactory=fmFactoryB
                              )
    gala = G.gala(settings)



    trainingInstance  = ngala.galaTrainingInstance(ragTrain, fOpTrain, edgeGt)
    gala.addTrainingInstance(trainingInstance)
    gala.train()


    for x in range(3,10):

        ragTest  = makeRag(imgs[x], showSeg=False)
        fOpTest, minVal, maxVal = makeFeatureOp(ragTest, imgs[x], minVal, maxVal)
        instance  = ngala.galaInstance(ragTest, fOpTest)
        edgeGt = makeEdgeGt(ragTest, gts[x])

        nodeRes = gala.predict(instance)

        pixelNodeRes = nrag.projectScalarNodeDataToPixels(ragTest,nodeRes,-1)
        vigra.segShow(imgs[x], pixelNodeRes)
        vigra.show()


        #for edge, uv in enumerate(ragTest.uvIds()):
        #    print(edge,edgeGt[edge],nodeRes[uv[0]]!=nodeRes[uv[1]])



test_mcgala()



from __future__ import absolute_import
from ._filters import *

import numpy
import scipy.ndimage.filters

import skimage.filters
import skimage.morphology

__all__ = []
for key in _filters.__dict__.keys():
    __all__.append(key)



def gaussianSmoothing(image, sigma, nSpatialDimensions=2):
    image = numpy.require(image, dtype='float32')
    if image.ndim == nSpatialDimensions:
        return scipy.ndimage.filters.gaussian_filter(image, sigma)
    elif image.ndim ==  nSpatialDimensions + 1:
        raise RuntimeError("not yer implemented")
    else:
        raise RuntimeError("image dimension does not match spatial dimension")

def gaussianGradientMagnitude(image, sigma, nSpatialDimensions=2):
    image = numpy.require(image, dtype='float32')
    if image.ndim == nSpatialDimensions:
        return scipy.ndimage.filters.gaussian_gradient_magnitude(image, sigma)
    elif image.ndim ==  nSpatialDimensions + 1:
        out = None
        nChannels = image.shape[image.ndim-1]
        for c in range(nChannels):
            cImage = image[...,c]
            gm = scipy.ndimage.filters.gaussian_gradient_magnitude(cImage, sigma)
            if out is None:
                out = gm
            else:
                out += gm
            out /= nChannels
        return out
    else:
        raise RuntimeError("image dimension does not match spatial dimension")
    

def affinitiesToProbability(affinities, edge_format=-1):
    ndim = affinities.ndim
    n_channels = affinities.shape[2]
    if ndim != 3 or n_channels != 2:
        raise RuntimeError("ndim must be 3 and n_channels must be 2")


    if edge_format == 1:
        ax = affinities[:, :, 0]
        ay = affinities[:, :, 1]

        ax_ = ax[0:-1,:   ]
        ay_ = ax[:   ,0:-1]


        axx = ax.copy()
        ayy = ay.copy()

        axx[1 :, :] += ax_
        ayy[:, 1 :] += ay_

    elif edge_format == -1:
        ax = affinities[:, :, 0]
        ay = affinities[:, :, 1]

        ax_ = ax[1:,:   ]
        ay_ = ax[:   ,1:]


        axx = ax.copy()
        ayy = ay.copy()

        axx[0:-1, :] += ax_
        ayy[:, 0:-1] += ay_
    
    else:
        raise RuntimeError("format must be in [1,-1]")


    return 1- (axx + ayy)/2.0



try :
    import vigra
    __has_vigra = True
except ImportError:
    __has_vigra = False


def diskMedian(img, radius):
    nimg = img.copy()
    oldMin = img.min()
    oldMax = img.max()
    nimg = numpy.require(nimg, dtype='float32')
    nimg -= oldMin
    nimg /= (oldMax - oldMin)
    nimg *= 255.0
    nimg = nimg.astype('uint8')
    disk  = skimage.morphology.disk(radius)
    r = skimage.filters.median(nimg, disk).astype('float32')/255.0
    r *= (oldMax - oldMin)
    r += oldMin
    return r




if __has_vigra:

    def hessianOfGaussianEigenvalues(image, sigma):

        imageShape = image.shape
        nDim = image.ndim 
        
        iamgeR = numpy.require(image, dtype='float32', requirements=['C'])
        imageT = iamgeR.T

        res = vigra.filters.hessianOfGaussianEigenvalues(imageT, sigma).view(numpy.ndarray).T
        res = numpy.moveaxis(res,0,-1)
        
        return numpy.require(res, requirements=['C'])



    def hessianOfGaussianStrongestEigenvalue(image, sigma):

        imageShape = image.shape
        nDim = image.ndim 
        
        iamgeR = numpy.require(image, dtype='float32', requirements=['C'])
        imageT = iamgeR.T

        res = vigra.filters.hessianOfGaussianEigenvalues(imageT, sigma)[:,:,0].view(numpy.ndarray).T

        return numpy.require(res, requirements=['C'])


from __future__ import print_function
import nifty
import numpy

def test_undirected_graph():

    g =  nifty.graph.UndirectedGraph(4)
    edges =  numpy.array([[0,1],[0,2],[0,3]],dtype='uint64')
    g.insertEdges(edges)

    edgeList = [e for e in g.edges()]
    assert edgeList == [0,1,2]

    nodeList = [e for e in g.nodes()]
    assert nodeList == [0,1,2,3]

    assert g.u(0) == 0
    assert g.v(0) == 1 
    assert g.u(1) == 0 
    assert g.v(1) == 2 
    assert g.u(2) == 0 
    assert g.v(2) == 3 




def make2x2Rag():

    labels = numpy.zeros(shape=[2,2],dtype='uint32')
    print(labels.shape)

    labels[0,0] = 0 
    labels[1,0] = 1 
    labels[0,1] = 0 
    labels[1,1] = 2 

    g =  nifty.graph.rag.explicitLabelsGridRag2D(labels)

    return g

def test_grid_rag():

    labels = numpy.zeros(shape=[2,2],dtype='uint32')
    print(labels.shape)

    labels[0,0] = 0 
    labels[1,0] = 1 
    labels[0,1] = 0 
    labels[1,1] = 2 

    g =  nifty.graph.rag.explicitLabelsGridRag2D(labels)
    weights = numpy.ones(g.numberOfEdges)*1
    obj = nifty.graph.multicut.multicutObjective(g, weights)


    greedy=obj.greedyAdditiveFactory().create(obj)
    visitor = obj.multicutVerboseVisitor()
    ret = greedy.optimizeWithVisitor(visitor=visitor)
    print("greedy",obj.evalNodeLabels(ret))




    assert g.numberOfNodes == 3
    assert g.numberOfEdges == 3

    insertWorked = True
    try:
        g.insertEdge(0,1)
    except:
        insertWorked = False
    assert insertWorked == False


test_grid_rag()



from __future__ import absolute_import
from . import _ufd as __ufd
from ._ufd import *


__all__ = []
for key in _ufd.__dict__.keys():
    __all__.append(key)
    try:
        __ufd.__dict__[key].__module__='nifty.ufd'
    except:
        pass

def ufd(size, dtype='uint64'):
    if dtype not in ['uint32','uint64']:
        raise RuntimeError("dtype must be 'uint32' or 'uint64'")
    if dtype == 'uint32':
        return Ufd_UInt32(int(size))
    elif dtype == 'uint64':
        return Ufd_UInt64(int(size))


def boost_ufd(elements, dtype='uint64'):
    if dtype not in ['uint32','uint64']:
        raise RuntimeError("dtype must be 'uint32' or 'uint64'")
    upper_bound = int(elements.max()) + 1
    if dtype == 'uint32':
        return BoostUfd_UInt32(elements, upper_bound)
    elif dtype == 'uint64':
        return BoostUfd_UInt64(elements, upper_bound)



from __future__ import print_function
import unittest
import numpy
import nifty
import nifty.graph


class TestShortesetPath(unittest.TestCase):
    def graphAndWeights(self):
        edges = numpy.array([
            [0,1],
            [1,2],
            [2,3],
            [3,4],
            [0,5],
            [4,5]
        ], dtype = 'uint64')
        g = nifty.graph.UndirectedGraph(6)
        g.insertEdges(edges)

        weights = numpy.zeros(len(edges), dtype = 'float32')
        weights[0]  = 1.
        weights[1]  = 1.
        weights[-1] = 5.

        return g, weights

    def testShortestPathDijkstraSingleTarget(self):
        g, weights = self.graphAndWeights()
        sp = nifty.graph.ShortestPathDijkstra(g)
        path = sp.runSingleSourceSingleTarget(weights, 0, 4)
        # shortest path 0 -> 4:
        # 0 - 1, 1 - 2, 3 - 4
        self.assertEqual(len(path), 5, str(len(path)))
        path.reverse()
        for ii in range(5):
            self.assertEqual(path[ii], ii, "%i, %i" % (path[ii], ii))

    def testShortestPathDijkstraSingleTargetParallel(self):
        g, weights = self.graphAndWeights()
        N = 50
        sources = N * [0]
        targets = N * [4]
        parallelPaths = nifty.graph.shortestPathSingleTargetParallel(
            g,
            weights.tolist(),
            sources,
            targets,
            returnNodes=True,
            numberOfThreads=4
        )

        for path in parallelPaths:
            # shortest path 0 -> 4:
            # 0 - 1, 1 - 2, 3 - 4
            self.assertEqual(len(path), 5, str(len(path)))
            path.reverse()
            for ii in range(5):
                self.assertEqual(path[ii], ii, "%i, %i" % (path[ii], ii))

    def testShortestPathDijkstraMultiTarget(self):
        g, weights = self.graphAndWeights()
        sp = nifty.graph.ShortestPathDijkstra(g)
        # we need to check 2 times to make sure that more than 1 runs work
        for _ in range(2):
            paths = sp.runSingleSourceMultiTarget(weights, 0, [4,5])
            self.assertEqual(len(paths), 2)
            # shortest path 0 -> 4:
            # 0 - 1, 1 - 2, 3 - 4
            path = paths[0]
            path.reverse()
            self.assertEqual(len(path), 5, str(len(path)))
            for ii in range(5):
                self.assertEqual(path[ii], ii, "%i, %i" % (path[ii], ii))

            # shortest path 0 -> 5:
            # 0 - 5
            path = paths[1]
            path.reverse()
            self.assertEqual(len(path) , 2, str(len(path)))
            self.assertEqual(path[0] , 0, str(path[0]))
            self.assertEqual(path[1] , 5, str(path[1]))

    def testShortestPathDijkstraMultiTargetParallel(self):
        g, weights = self.graphAndWeights()
        N = 50
        sources = N*[0]
        targets = [[4,5] for _ in range(N)]

        for _ in range(2):
            parallelPaths = nifty.graph.shortestPathMultiTargetParallel(
                g,
                weights,
                sources,
                targets,
                returnNodes=True,
                numberOfThreads=5
            )
            for paths in parallelPaths:
                self.assertEqual(len(paths), 2)
                # shortest path 0 -> 4:
                # 0 - 1, 1 - 2, 3 - 4
                path = paths[0]
                path.reverse()
                self.assertEqual(len(path), 5, str(len(path)))
                for ii in range(5):
                    self.assertEqual(path[ii] , ii, "%i, %i" % (path[ii], ii))

                # shortest path 0 -> 5:
                # 0 - 5
                path = paths[1]
                path.reverse()
                self.assertEqual(len(path) , 2, str(len(path)))
                self.assertEqual(path[0] , 0, str(path[0]))
                self.assertEqual(path[1] , 5, str(path[1]))

    def testShortestPathInvalid(self):
        edges = numpy.array([
            [0,1],
            [2,3]
        ], dtype = 'uint64')
        g = nifty.graph.UndirectedGraph(4)
        g.insertEdges(edges)
        sp = nifty.graph.ShortestPathDijkstra(g)
        weights = [1.,1.]
        path = sp.runSingleSourceSingleTarget(weights, 0, 3)
        self.assertTrue(not path) # make sure that the path is invalid


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np
import nifty.graph.rag as nrag


# Just a quick and dirty test to check that the functions can be called
class TestAccumulate(unittest.TestCase):
    shape_2d = 2 * (128,)
    shape_3d = 3 * (64,)

    def test_accumulate_mean_and_length_2d(self):
        labels = np.random.randint(0, 100, size=self.shape_2d, dtype='uint32')
        rag = nrag.gridRag(labels, numberOfLabels=100)
        data = np.random.random_sample(self.shape_2d).astype('float32')
        res = nrag.accumulateEdgeMeanAndLength(rag, data)
        self.assertTrue(np.sum(res) != 0)

    def test_accumulate_mean_and_length_3d(self):
        labels = np.random.randint(0, 100, size=self.shape_3d, dtype='uint32')
        rag = nrag.gridRag(labels, numberOfLabels=100)
        data = np.random.random_sample(self.shape_3d).astype('float32')
        res = nrag.accumulateEdgeMeanAndLength(rag, data)
        self.assertTrue(np.sum(res) != 0)



if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np
import nifty.tools as nt


class TestEdgeMapping(unittest.TestCase):

    def test_edge_mapping_toy(self):
        uv_ids = np.array([[0, 1],
                           [0, 2],
                           [1, 2],
                           [2, 3],
                           [2, 4],
                           [2, 5],
                           [2, 6],
                           [3, 4],
                           [3, 6],
                           [4, 5]], dtype='int64')

        # node labeling           0  1  2  3  4  5  6
        node_labeling = np.array([0, 1, 0, 2, 1, 2, 3], dtype='uint64')
        self.assertEqual(len(node_labeling), len(np.unique(uv_ids)))
        edge_mapping = nt.EdgeMapping(uv_ids, node_labeling)

        new_uv_ids = edge_mapping.newUvIds()
        new_uv_ids_exp = np.array([[0, 1],
                                   [0, 2],
                                   [0, 3],
                                   [1, 2],
                                   [2, 3]], dtype='uint64')

        self.assertEqual(new_uv_ids.shape, new_uv_ids_exp.shape)
        self.assertTrue((new_uv_ids == new_uv_ids_exp).all())

        # test edge mappings
        # mapping
        # edge 0: 0 -> 1 == 0 -> 1 : 0
        # edge 1: 0 -> 2 == 0 -> 0 : Null
        # edge 2: 1 -> 2 == 1 -> 0 : 0
        # edge 3: 2 -> 3 == 0 -> 2 : 1
        # edge 4: 2 -> 4 == 0 -> 1 : 0
        # edge 5: 2 -> 5 == 0 -> 2 : 1
        # edge 6: 2 -> 6 == 0 -> 3 : 2
        # edge 7: 3 -> 4 == 2 -> 1 : 3
        # edge 8: 3 -> 6 == 2 -> 3 : 4
        # edge 9: 4 -> 5 == 1 -> 2 : 3

        edge_values = np.ones(len(uv_ids), dtype='float32')
        # sum mapping
        new_values = edge_mapping.mapEdgeValues(edge_values, "sum")
        new_values_exp = np.array([3, 2, 1, 2, 1], dtype='float32')
        self.assertEqual(new_values.shape, new_values_exp.shape)
        self.assertTrue(np.allclose(new_values, new_values_exp))

        # mean mapping
        new_values = edge_mapping.mapEdgeValues(edge_values, "mean")
        new_values_exp = np.array([1, 1, 1, 1, 1], dtype='float32')
        self.assertEqual(new_values.shape, new_values_exp.shape)
        self.assertTrue(np.allclose(new_values, new_values_exp))

        edge_values = np.ones(len(uv_ids), dtype='float32')
        edge_values[::2] = 0
        # min mapping
        # edges:  0 1 2 3 4 5 6 7 8 9
        # values: 0 1 0 1 0 1 0 1 0 1
        new_values = edge_mapping.mapEdgeValues(edge_values, "min")
        new_values_exp = np.array([0, 1, 0, 1, 0], dtype='float32')
        self.assertEqual(new_values.shape, new_values_exp.shape)
        self.assertTrue(np.allclose(new_values, new_values_exp))

        # max mapping
        # edges:  0 1 2 3 4 5 6 7 8 9
        # values: 0 1 0 1 0 1 0 1 0 1
        new_values = edge_mapping.mapEdgeValues(edge_values, "max")
        new_values_exp = np.array([0, 1, 0, 1, 0], dtype='float32')
        self.assertEqual(new_values.shape, new_values_exp.shape)
        self.assertTrue(np.allclose(new_values, new_values_exp))


if __name__ == '__main__':
    unittest.main()



from __future__ import print_function


import unittest
import numpy

import nifty
import nifty.graph
import nifty.graph.agglo
nagglo = nifty.graph.agglo


class TestGASP(unittest.TestCase):
    def setUp(self):
        # Create a small graph:
        self.g = g = nifty.graph.UndirectedGraph(4)
        edges = numpy.array([
            [0, 1], [0, 2], [0, 3],
            [1, 3],
            [2, 3]
        ], dtype='uint64')
        g.insertEdges(edges)

        self.edgeIndicators = numpy.array([-10, -2, 6, 3, 11], dtype='float32')

    def test_gasp_average(self):
        clusterPolicy = nagglo.get_GASP_policy(
            graph=self.g,
            signed_edge_weights=self.edgeIndicators,
            linkage_criteria='mean',
            add_cannot_link_constraints=False)

        agglomerativeClustering = nagglo.agglomerativeClustering(clusterPolicy)
        agglomerativeClustering.run()
        seg = agglomerativeClustering.result()
        self.assertTrue(seg[0] != seg[1] and seg[1] == seg[2] and seg[2] == seg[3])

    def test_gasp_abs_max(self):
        clusterPolicy = nagglo.get_GASP_policy(
            graph=self.g,
            signed_edge_weights=self.edgeIndicators,
            linkage_criteria='abs_max')

        agglomerativeClustering = nagglo.agglomerativeClustering(clusterPolicy)
        agglomerativeClustering.run()
        seg = agglomerativeClustering.result()
        self.assertTrue(seg[0] != seg[1] and seg[0] == seg[2] and seg[0] == seg[3])


    def test_gasp_sum(self):
        clusterPolicy = nagglo.get_GASP_policy(
            graph=self.g,
            signed_edge_weights=self.edgeIndicators,
            linkage_criteria='sum',
            add_cannot_link_constraints=False)

        agglomerativeClustering = nagglo.agglomerativeClustering(clusterPolicy)
        agglomerativeClustering.run()
        seg = agglomerativeClustering.result().tolist()
        self.assertTrue(seg[0] != seg[1] and seg[0] == seg[2] and seg[0] == seg[3])



if __name__ == '__main__':
    unittest.main()